#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
"""Provides the DynaMOSA test-generation strategy."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

import networkx as nx
from networkx.drawing.nx_pydot import to_pydot
from ordered_set import OrderedSet

import pynguin.configuration as config
import pynguin.coverage.branchgoals as bg
import pynguin.utils.statistics.statistics as stat
from pynguin.ga.operators.ranking.crowdingdistance import (
    fast_epsilon_dominance_assignment,
)
from pynguin.generation.algorithms.abstractmosastrategy import AbstractMOSATestStrategy
from pynguin.utils.statistics.runtimevariable import RuntimeVariable

if TYPE_CHECKING:
    import pynguin.ga.computations as ff
    import pynguin.ga.testcasechromosome as tcc
    import pynguin.ga.testsuitechromosome as tsc
    from pynguin.generation.algorithms.archive import CoverageArchive
    from pynguin.testcase.execution import KnownData


class DynaMOSATestStrategy(AbstractMOSATestStrategy):
    """Implements the Dynamic Many-Objective Sorting Algorithm DynaMOSA."""

    _logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        super().__init__()
        self._goals_manager: _GoalsManager

    def generate_tests(self) -> tsc.TestSuiteChromosome:
        self.before_search_start()
        self._goals_manager = _GoalsManager(
            self._test_case_fitness_functions,
            self._archive,
            self.executor.tracer.get_known_data(),
        )
        self._number_of_goals = len(self._test_case_fitness_functions)
        stat.set_output_variable_for_runtime_variable(
            RuntimeVariable.Goals, self._number_of_goals
        )

        self._population = self._get_random_population()
        self._goals_manager.update(self._population)

        # Calculate dominance ranks and crowding distance
        fronts = self._ranking_function.compute_ranking_assignment(
            self._population, self._goals_manager.current_goals
        )
        for i in range(fronts.get_number_of_sub_fronts()):
            fast_epsilon_dominance_assignment(
                fronts.get_sub_front(i), self._goals_manager.current_goals
            )

        self.before_first_search_iteration(
            self.create_test_suite(self._archive.solutions)
        )
        while self.resources_left() and len(self._archive.uncovered_goals) > 0:
            self.evolve()
            self.after_search_iteration(self.create_test_suite(self._archive.solutions))

        self.after_search_finish()
        return self.create_test_suite(
            self._archive.solutions
            if len(self._archive.solutions) > 0
            else self._get_best_individuals()
        )

    def evolve(self) -> None:
        """Runs one evolution step."""
        offspring_population: list[
            tcc.TestCaseChromosome
        ] = self._breed_next_generation()

        # Create union of parents and offspring
        union: list[tcc.TestCaseChromosome] = []
        union.extend(self._population)
        union.extend(offspring_population)

        # Ranking the union
        self._logger.debug("Union Size = %d", len(union))
        # Ranking the union using the best rank algorithm
        fronts = self._ranking_function.compute_ranking_assignment(
            union, self._goals_manager.current_goals
        )

        # Form the next population using “preference sorting and non-dominated
        # sorting” on the updated set of goals
        remain = max(
            config.configuration.search_algorithm.population,
            len(fronts.get_sub_front(0)),
        )
        index = 0
        self._population.clear()

        # Obtain the first front
        front = fronts.get_sub_front(index)

        while remain > 0 and remain >= len(front) != 0:
            # Assign crowding distance to individuals
            fast_epsilon_dominance_assignment(front, self._goals_manager.current_goals)
            # Add the individuals of this front
            self._population.extend(front)
            # Decrement remain
            remain -= len(front)
            # Obtain the next front
            index += 1
            if remain > 0:
                front = fronts.get_sub_front(index)

        # Remain is less than len(front[index]), insert only the best one
        if remain > 0 and len(front) != 0:
            fast_epsilon_dominance_assignment(front, self._goals_manager.current_goals)
            front.sort(key=lambda t: t.distance, reverse=True)
            for k in range(remain):
                self._population.append(front[k])

        self._goals_manager.update(self._population)


class _GoalsManager:
    """Manages goals and provides dynamically selected ones for the generation."""

    _logger = logging.getLogger(__name__)

    def __init__(
        self,
        fitness_functions: OrderedSet[ff.FitnessFunction],
        archive: CoverageArchive,
        known_data: KnownData,
    ) -> None:
        self._archive = archive
        branch_fitness_functions: OrderedSet[
            bg.BranchCoverageTestFitness
        ] = OrderedSet()
        for fit in fitness_functions:
            assert isinstance(fit, bg.BranchCoverageTestFitness)
            branch_fitness_functions.add(fit)
        self._graph = _BranchFitnessGraph(branch_fitness_functions, known_data)
        self._current_goals: OrderedSet[
            bg.BranchCoverageTestFitness
        ] = self._graph.root_branches
        self._archive.add_goals(self._current_goals)

    @property
    def current_goals(self) -> OrderedSet[ff.FitnessFunction]:
        """Provides the set of current goals.

        Returns:
            The set of current goals
        """
        return self._current_goals

    def update(self, solutions: list[tcc.TestCaseChromosome]) -> None:
        """Updates the information on the current goals from the found solutions.

        Args:
            solutions: The previously found solutions
        """
        # We must keep iterating, as long as new goals are added.
        new_goals_added = True
        while new_goals_added:
            self._archive.update(solutions)
            covered = self._archive.covered_goals
            new_goals = OrderedSet()
            new_goals_added = False
            for old_goal in self._current_goals:
                if old_goal in covered:
                    children = self._graph.get_structural_children(
                        cast(bg.BranchCoverageTestFitness, old_goal)
                    )
                    for child in children:
                        if child not in self._current_goals and child not in covered:
                            new_goals.add(child)
                            new_goals_added = True
                else:
                    new_goals.add(old_goal)
            self._current_goals = new_goals
            self._archive.add_goals(self._current_goals)
        self._logger.debug("current goals after update: %s", self._current_goals)


class _BranchFitnessGraph:
    """Best effort re-implementation of EvoSuite's BranchFitnessGraph.

    Arranges the fitness functions for all branches according to their control
    dependencies in the CDG. Each node represents a fitness function. A directed edge
    (u -> v) states that fitness function v should be added for consideration
    only when fitness function u has been covered."""

    def __init__(
        self,
        fitness_functions: OrderedSet[bg.BranchCoverageTestFitness],
        known_data: KnownData,
    ):
        self._graph = nx.DiGraph()
        # Branch less code objects and branches that are not control dependent on other
        # branches.
        self._root_branches = OrderedSet()
        self._build_graph(fitness_functions, known_data)

    def _build_graph(
        self,
        fitness_functions: OrderedSet[bg.BranchCoverageTestFitness],
        known_data: KnownData,
    ):
        """Construct the actual graph from the given fitness functions."""
        for fitness in fitness_functions:
            self._graph.add_node(fitness)

        for fitness in fitness_functions:
            if fitness.goal.is_branchless_code_object:
                self._root_branches.add(fitness)
                continue
            assert fitness.goal.is_branch
            branch_goal = cast(bg.BranchGoal, fitness.goal)
            predicate_meta_data = known_data.existing_predicates[
                branch_goal.predicate_id
            ]
            code_object_meta_data = known_data.existing_code_objects[
                predicate_meta_data.code_object_id
            ]
            if code_object_meta_data.cdg.is_control_dependent_on_root(
                predicate_meta_data.node
            ):
                self._root_branches.add(fitness)

            dependencies = code_object_meta_data.cdg.get_control_dependencies(
                predicate_meta_data.node
            )
            for dependency in dependencies:
                goal = bg.BranchGoal(
                    predicate_meta_data.code_object_id,
                    dependency.predicate_id,
                    dependency.branch_value,
                )
                dependent_ff = self._goal_to_fitness_function(fitness_functions, goal)
                self._graph.add_edge(dependent_ff, fitness)

        # Sanity check
        assert {n for n in self._graph.nodes if self._graph.in_degree(n) == 0}.issubset(
            self._root_branches
        ), "Root branches cannot depend on other branches."

    @property
    def dot(self):
        """Return DOT representation of this graph."""
        dot = to_pydot(self._graph)
        return dot.to_string()

    @property
    def root_branches(self) -> OrderedSet[bg.BranchCoverageTestFitness]:
        """Return the root branches, i.e., the fitness functions that have
        no preconditions."""
        return OrderedSet(self._root_branches)

    @staticmethod
    def _goal_to_fitness_function(
        search_in: OrderedSet[bg.BranchCoverageTestFitness], goal: bg.BranchGoal
    ) -> bg.BranchCoverageTestFitness:
        """Little helper to find the fitness function associated with a certain goal.

        Args:
            search_in: The list to search in
            goal: The goal to search for

        Returns:
            The found fitness function.
        """
        for fitness in search_in:
            if fitness.goal == goal:
                return fitness
        raise RuntimeError(f"Could not find fitness function for goal: {goal}")

    def get_structural_children(
        self, fitness_function: bg.BranchCoverageTestFitness
    ) -> OrderedSet[bg.BranchCoverageTestFitness]:
        """Get the fitness functions that are structural children of the given
        fitness function.

        Args:
            fitness_function: The fitness function whose structural children should be
            returned.

        Returns:
            The structural children fitness functions of the given fitness function.
        """
        return OrderedSet(self._graph.successors(fitness_function))



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an assertion visitor to transform assertions to AST."""
from __future__ import annotations

import ast
from typing import Any

import pynguin.assertion.assertion as ass
import pynguin.configuration as config
import pynguin.testcase.variablereference as vr
import pynguin.utils.ast_util as au
import pynguin.utils.namingscope as ns
import pynguin.utils.type_utils as tu


class AssertionToAstVisitor(ass.AssertionVisitor):
    """An assertion visitor that transforms assertions into AST nodes."""

    def __init__(
        self,
        variable_names: ns.AbstractNamingScope,
        module_aliases: ns.AbstractNamingScope,
        common_modules: set[str],
    ):
        """Create a new assertion visitor.

        Args:
            variable_names: the naming scope that is used to resolve the names
                            of the variables used in the assertions.
            module_aliases: the naming scope that is used to resolve the aliases of the
                            modules used in the assertions.
            common_modules: the set of common modules that are used. Modules may be
                            added when transforming the assertions.
        """
        self._common_modules = common_modules
        self._module_aliases = module_aliases
        self._variable_names = variable_names
        self._nodes: list[ast.stmt] = []

    @property
    def nodes(self) -> list[ast.stmt]:
        """Provides the ast nodes generated by this visitor.

        Returns:
            the ast nodes generated by this visitor.
        """
        return self._nodes

    def visit_float_assertion(self, assertion: ass.FloatAssertion) -> None:
        """Creates an assertion of form
        "assert float_0 == pytest.approx(1, rel=0.01, abs=0.01)"

        Args:
            assertion: the assertion that is visited.

        """
        left = au.create_full_name(
            self._variable_names, self._module_aliases, assertion.source, load=True
        )
        comp = self._construct_float_comparator(au.create_ast_constant(assertion.value))
        self._nodes.append(
            au.create_ast_assert(au.create_ast_compare(left, ast.Eq(), comp))
        )

    def visit_not_none_assertion(self, assertion: ass.NotNoneAssertion) -> None:
        """
        Creates an assertion of form "assert var0 is None" or "assert var0 is not None".

        Args:
            assertion: the assertion that is visited.
        """
        self._nodes.append(
            self._create_constant_assert(assertion.source, ast.IsNot(), None)
        )

    def visit_object_assertion(self, assertion: ass.ObjectAssertion) -> None:
        """
        Creates an assertion of form "assert var0 == value" or "assert var0 is False",
        if the value is a bool.

        Args:
            assertion: the assertion that is visited.
        """
        if isinstance(assertion.object, (bool, type(None))):
            self._nodes.append(
                self._create_constant_assert(
                    assertion.source, ast.Is(), assertion.object
                )
            )
        else:
            comp = self._create_assertable_object(assertion.object)
            left = au.create_full_name(
                self._variable_names, self._module_aliases, assertion.source, load=True
            )
            self._nodes.append(
                au.create_ast_assert(au.create_ast_compare(left, ast.Eq(), comp))
            )

    def _create_assertable_object(self, value: Any):
        """Recursively constructs an assertable object. Must handle all cases that
        `pynguin.utils.type_utils.is_assertable` accepts.

        Args:
            value: The value that should be generated.

        Returns:
            An assertion representation of the given value.

        Raises:
            AssertionError: If we encounter an object that we can't construct.
        """
        tp_ = type(value)
        if tu.is_enum(tp_):
            enum_attr = self._construct_enum_attr(value)
            return au.create_ast_attribute(value.name, enum_attr)
        if tu.is_primitive_type(tp_) or tu.is_none_type(tp_):
            return au.create_ast_constant(value)
        if tu.is_set(value) or tu.is_list(value) or tu.is_tuple(value):
            elements = [self._create_assertable_object(v) for v in value]
            if tu.is_set(value):
                return au.create_ast_set(elements)
            if tu.is_list(value):
                return au.create_ast_list(elements)
            return au.create_ast_tuple(elements)
        if tu.is_dict(value):
            keys = [self._create_assertable_object(v) for v in value]
            values = [self._create_assertable_object(v) for v in value.values()]
            return au.create_ast_dict(keys, values)
        raise AssertionError(f"Cannot create assertion object of type {type(value)}")

    def _create_constant_assert(
        self, var: vr.Reference, operator: ast.cmpop, value: Any
    ) -> ast.Assert:
        left = au.create_full_name(
            self._variable_names, self._module_aliases, var, load=True
        )
        comp = au.create_ast_constant(value)
        return au.create_ast_assert(au.create_ast_compare(left, operator, comp))

    def _construct_float_comparator(self, comp):
        self._common_modules.add("pytest")
        float_precision = config.configuration.test_case_output.float_precision
        func = au.create_ast_attribute("approx", au.create_ast_name("pytest"))
        keywords = [
            au.create_ast_keyword("abs", au.create_ast_constant(float_precision)),
            au.create_ast_keyword("rel", au.create_ast_constant(float_precision)),
        ]
        comp_float = au.create_ast_call(func, [comp], keywords)
        return comp_float

    def _construct_enum_attr(self, value) -> ast.Attribute:
        module = self._module_aliases.get_name(value.__class__.__module__)
        enum_name = value.__class__.__name__
        return au.create_ast_attribute(enum_name, au.create_ast_name(module))

    def visit_collection_length_assertion(
        self, assertion: ass.CollectionLengthAssertion
    ) -> None:
        left = au.create_ast_call(
            au.create_ast_name("len"),
            [
                au.create_full_name(
                    self._variable_names, self._module_aliases, assertion.source, True
                )
            ],
            {},
        )
        self._nodes.append(
            au.create_ast_assert(
                au.create_ast_compare(
                    left, ast.Eq(), au.create_ast_constant(assertion.length)
                )
            )
        )



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an interface for a stopping condition of the algorithm."""
from __future__ import annotations

import time
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

import pynguin.configuration as config
import pynguin.generation.searchobserver as so
from pynguin.testcase.execution import ExecutionObserver

if TYPE_CHECKING:
    import pynguin.ga.testsuitechromosome as tsc
    import pynguin.testcase.statement as stmt
    import pynguin.testcase.testcase as tc
    from pynguin.testcase.execution import ExecutionContext, ExecutionResult


class StoppingCondition(so.SearchObserver, ExecutionObserver, metaclass=ABCMeta):
    """Provides an interface for a stopping condition of the algorithm."""

    def __init__(self, observes_execution: bool = False):
        self._observes_execution = observes_execution

    @abstractmethod
    def current_value(self) -> int:
        """Provide how much of the budget we have used.

        Returns:
            The current value of the budget
        """

    @abstractmethod
    def limit(self) -> int:
        """Get upper limit of resources.

        Returns:
            The limit  # noqa: DAR202
        """

    @abstractmethod
    def is_fulfilled(self) -> bool:
        """Returns whether the condition is fulfilled, thus the algorithm should stop

        Returns:
            True if the condition is fulfilled, False otherwise  # noqa: DAR202
        """

    @abstractmethod
    def reset(self) -> None:
        """Reset everything."""

    @abstractmethod
    def set_limit(self, limit: int) -> None:
        """Sets new upper limit of resources.

        Args:
            limit: The new upper limit
        """

    def __str__(self):
        return f"{self.__class__.__name__}: {self.current_value()} / {self.limit()}"

    @property
    def observes_execution(self) -> bool:
        """Should this observer be attached to the executor?"""
        return self._observes_execution

    def before_test_case_execution(self, test_case: tc.TestCase):
        pass

    def after_test_case_execution(
        self, test_case: tc.TestCase, result: ExecutionResult
    ):
        pass

    def before_statement_execution(
        self, statement: stmt.Statement, exec_ctx: ExecutionContext
    ):
        pass

    def after_statement_execution(
        self,
        statement: stmt.Statement,
        exec_ctx: ExecutionContext,
        exception: Exception | None = None,
    ) -> None:
        pass

    def before_search_start(self, start_time_ns: int) -> None:
        pass

    def before_first_search_iteration(self, initial: tsc.TestSuiteChromosome) -> None:
        pass

    def after_search_iteration(self, best: tsc.TestSuiteChromosome) -> None:
        pass

    def after_search_finish(self) -> None:
        pass


class MaxIterationsStoppingCondition(StoppingCondition):
    """A stopping condition that checks the maximum number of test cases."""

    def __init__(self):
        super().__init__()
        self._num_iterations = 0
        self._max_iterations = config.configuration.stopping.maximum_iterations

    def current_value(self) -> int:
        return self._num_iterations

    def limit(self) -> int:
        return self._max_iterations

    def is_fulfilled(self) -> bool:
        return self._num_iterations >= self._max_iterations

    def reset(self) -> None:
        self._num_iterations = 0

    def set_limit(self, limit: int) -> None:
        self._max_iterations = limit

    def before_search_start(self, start_time_ns: int) -> None:
        self._num_iterations = 0

    def after_search_iteration(self, best: tsc.TestSuiteChromosome) -> None:
        self._num_iterations += 1


class MaxTestExecutionsStoppingCondition(StoppingCondition):
    """A stopping condition that checks the maximum number of test case executions."""

    def __init__(self):
        super().__init__(observes_execution=True)
        self._num_executed_tests = 0
        self._max_test_executions = (
            config.configuration.stopping.maximum_test_executions
        )

    def current_value(self) -> int:
        return self._num_executed_tests

    def limit(self) -> int:
        return self._max_test_executions

    def is_fulfilled(self) -> bool:
        return self._num_executed_tests >= self._max_test_executions

    def reset(self) -> None:
        self._num_executed_tests = 0

    def set_limit(self, limit: int) -> None:
        self._max_test_executions = limit

    def before_search_start(self, start_time_ns: int) -> None:
        self._num_executed_tests = 0

    def after_test_case_execution(
        self, test_case: tc.TestCase, result: ExecutionResult
    ):
        self._num_executed_tests += 1


class MaxStatementExecutionsStoppingCondition(StoppingCondition):
    """A stopping condition that checks the maximum number of executed statements."""

    def __init__(self):
        super().__init__(observes_execution=True)
        self._num_executed_statements = 0
        self._max_executed_statements = (
            config.configuration.stopping.maximum_statement_executions
        )

    def current_value(self) -> int:
        return self._num_executed_statements

    def limit(self) -> int:
        return self._max_executed_statements

    def is_fulfilled(self) -> bool:
        return self._num_executed_statements >= self._max_executed_statements

    def reset(self) -> None:
        self._num_executed_statements = 0

    def set_limit(self, limit: int) -> None:
        self._max_executed_statements = limit

    def before_search_start(self, start_time_ns: int) -> None:
        self._num_executed_statements = 0

    def after_statement_execution(
        self,
        statement: stmt.Statement,
        exec_ctx: ExecutionContext,
        exception: Exception | None = None,
    ) -> None:
        self._num_executed_statements += 1


class MaxTimeStoppingCondition(StoppingCondition):
    """Stop search after a predefined amount of time."""

    def __init__(self):
        super().__init__()
        self._max_seconds = config.configuration.stopping.budget
        self._start_time = 0

    def current_value(self) -> int:
        return (time.time_ns() - self._start_time) // 1_000_000_000

    def limit(self) -> int:
        return self._max_seconds

    def is_fulfilled(self) -> bool:
        return ((time.time_ns() - self._start_time) / 1_000_000_000) > self._max_seconds

    def reset(self) -> None:
        self._start_time = time.time_ns()

    def set_limit(self, limit: int) -> None:
        self._max_seconds = limit

    def before_search_start(self, start_time_ns: int) -> None:
        self._start_time = start_time_ns



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides factories for the generation algorithm."""
from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Callable, Generic, TypeVar

from ordered_set import OrderedSet

import pynguin.configuration as config
import pynguin.coverage.branchgoals as bg
import pynguin.ga.chromosome as chrom
import pynguin.ga.computations as ff
import pynguin.ga.testcasechromosomefactory as tccf
import pynguin.ga.testcasefactory as tcf
import pynguin.ga.testsuitechromosome as tsc
import pynguin.ga.testsuitechromosomefactory as tscf
import pynguin.generation.algorithms.archive as arch
import pynguin.generation.searchobserver as so
import pynguin.testcase.testfactory as tf
import pynguin.utils.statistics.statisticsobserver as sso
from pynguin.ga.operators.crossover.singlepointrelativecrossover import (
    SinglePointRelativeCrossOver,
)
from pynguin.ga.operators.ranking.rankingfunction import RankBasedPreferenceSorting
from pynguin.ga.operators.selection.rankselection import RankSelection
from pynguin.ga.operators.selection.selection import SelectionFunction
from pynguin.ga.operators.selection.tournamentselection import TournamentSelection
from pynguin.generation.algorithms.dynamosastrategy import DynaMOSATestStrategy
from pynguin.generation.algorithms.mioteststrategy import MIOTestStrategy
from pynguin.generation.algorithms.mosastrategy import MOSATestStrategy
from pynguin.generation.algorithms.randomsearchstrategy import (
    RandomTestCaseSearchStrategy,
    RandomTestSuiteSearchStrategy,
)
from pynguin.generation.algorithms.randomteststrategy import RandomTestStrategy
from pynguin.generation.algorithms.wholesuiteteststrategy import WholeSuiteTestStrategy
from pynguin.generation.stoppingconditions.stoppingcondition import (
    MaxIterationsStoppingCondition,
    MaxStatementExecutionsStoppingCondition,
    MaxTestExecutionsStoppingCondition,
    MaxTimeStoppingCondition,
)
from pynguin.setup.testcluster import FilteredTestCluster, TestCluster
from pynguin.testcase.execution import TestCaseExecutor
from pynguin.utils.exceptions import ConfigurationException

if TYPE_CHECKING:
    import pynguin.ga.chromosomefactory as cf
    from pynguin.ga.operators.crossover.crossover import CrossOverFunction
    from pynguin.ga.operators.ranking.rankingfunction import RankingFunction
    from pynguin.generation.algorithms.testgenerationstrategy import (
        TestGenerationStrategy,
    )
    from pynguin.generation.stoppingconditions.stoppingcondition import (
        StoppingCondition,
    )

C = TypeVar("C", bound=chrom.Chromosome)  # pylint: disable=invalid-name


class GenerationAlgorithmFactory(Generic[C], metaclass=ABCMeta):
    """A generic generation algorithm factory."""

    _logger = logging.getLogger(__name__)

    # pylint:disable=line-too-long
    _stopping_conditions: dict[
        config.StoppingCondition, Callable[[], StoppingCondition]
    ] = {
        config.StoppingCondition.MAX_ITERATIONS: MaxIterationsStoppingCondition,
        config.StoppingCondition.MAX_TEST_EXECUTIONS: MaxTestExecutionsStoppingCondition,
        config.StoppingCondition.MAX_STATEMENT_EXECUTIONS: MaxStatementExecutionsStoppingCondition,
        config.StoppingCondition.MAX_TIME: MaxTimeStoppingCondition,
    }

    def get_stopping_condition(self) -> StoppingCondition:
        """Instantiates the stopping condition depending on the configuration settings.

        Returns:
            A stopping condition
        """
        stopping_condition = config.configuration.stopping.stopping_condition
        self._logger.info("Using stopping condition: %s", stopping_condition)
        if stopping_condition in self._stopping_conditions:
            return self._stopping_conditions[stopping_condition]()
        self._logger.warning("Unknown stopping condition: %s", stopping_condition)
        return MaxTimeStoppingCondition()

    @abstractmethod
    def get_search_algorithm(self) -> TestGenerationStrategy:
        """Initialises and sets up the test-generation strategy to use.

        Returns:
            A fully configured test-generation strategy  # noqa: DAR202
        """


# pylint: disable=unsubscriptable-object, too-few-public-methods
class TestSuiteGenerationAlgorithmFactory(
    GenerationAlgorithmFactory[tsc.TestSuiteChromosome]
):
    """A factory for a search algorithm generating test-suites."""

    _strategies: dict[config.Algorithm, Callable[[], TestGenerationStrategy]] = {
        config.Algorithm.DYNAMOSA: DynaMOSATestStrategy,
        config.Algorithm.MIO: MIOTestStrategy,
        config.Algorithm.MOSA: MOSATestStrategy,
        config.Algorithm.RANDOM: RandomTestStrategy,
        config.Algorithm.RANDOM_TEST_SUITE_SEARCH: RandomTestSuiteSearchStrategy,
        config.Algorithm.RANDOM_TEST_CASE_SEARCH: RandomTestCaseSearchStrategy,
        config.Algorithm.WHOLE_SUITE: WholeSuiteTestStrategy,
    }

    _selections: dict[config.Selection, Callable[[], SelectionFunction]] = {
        config.Selection.TOURNAMENT_SELECTION: TournamentSelection,
        config.Selection.RANK_SELECTION: RankSelection,
    }

    def __init__(self, executor: TestCaseExecutor, test_cluster: TestCluster):
        self._executor = executor
        self._test_cluster = test_cluster

    def _get_chromosome_factory(
        self, strategy: TestGenerationStrategy
    ) -> cf.ChromosomeFactory:
        """Provides a chromosome factory.

        Args:
            strategy: The strategy that is currently configured.

        Returns:
            A chromosome factory
        """
        # TODO add conditional returns/other factories here
        test_case_factory: tcf.TestCaseFactory = tcf.RandomLengthTestCaseFactory(
            strategy.test_factory
        )
        if config.configuration.seeding.initial_population_seeding:
            self._logger.info("Using population seeding")
            test_case_factory = tcf.SeededTestCaseFactory(
                test_case_factory, strategy.test_factory
            )
        test_case_chromosome_factory: cf.ChromosomeFactory = (
            tccf.TestCaseChromosomeFactory(
                strategy.test_factory,
                test_case_factory,
                strategy.test_case_fitness_functions,
            )
        )
        if config.configuration.seeding.seed_from_archive:
            self._logger.info("Using archive seeding")
            test_case_chromosome_factory = tccf.ArchiveReuseTestCaseChromosomeFactory(
                test_case_chromosome_factory, strategy.archive
            )
        if config.configuration.algorithm in (
            config.Algorithm.DYNAMOSA,
            config.Algorithm.MIO,
            config.Algorithm.MOSA,
            config.Algorithm.RANDOM_TEST_CASE_SEARCH,
        ):
            return test_case_chromosome_factory
        return tscf.TestSuiteChromosomeFactory(
            test_case_chromosome_factory,
            strategy.test_suite_fitness_functions,
            strategy.test_suite_coverage_functions,
        )

    def get_search_algorithm(self) -> TestGenerationStrategy:
        """Initialises and sets up the test-generation strategy to use.

        Returns:
            A fully configured test-generation strategy
        """
        strategy = self._get_generation_strategy()
        strategy.branch_goal_pool = bg.BranchGoalPool(
            self._executor.tracer.get_known_data()
        )
        strategy.test_case_fitness_functions = self._get_test_case_fitness_functions(
            strategy
        )
        strategy.test_suite_fitness_functions = self._get_test_suite_fitness_functions()
        strategy.test_suite_coverage_functions = (
            self._get_test_suite_coverage_functions()
        )
        strategy.archive = self._get_archive(strategy)

        strategy.executor = self._executor
        strategy.test_cluster = self._get_test_cluster(strategy)
        strategy.test_factory = self._get_test_factory(strategy)
        chromosome_factory = self._get_chromosome_factory(strategy)
        strategy.chromosome_factory = chromosome_factory

        selection_function = self._get_selection_function()
        selection_function.maximize = False
        strategy.selection_function = selection_function

        stopping_condition = self.get_stopping_condition()
        strategy.stopping_condition = stopping_condition
        strategy.add_search_observer(stopping_condition)
        if stopping_condition.observes_execution:
            self._executor.add_observer(stopping_condition)
        strategy.add_search_observer(so.LogSearchObserver())
        strategy.add_search_observer(sso.SequenceStartTimeObserver())
        strategy.add_search_observer(sso.IterationObserver())
        strategy.add_search_observer(sso.BestIndividualObserver())

        crossover_function = self._get_crossover_function()
        strategy.crossover_function = crossover_function

        ranking_function = self._get_ranking_function()
        strategy.ranking_function = ranking_function

        return strategy

    @classmethod
    def _get_generation_strategy(cls) -> TestGenerationStrategy:
        """Provides a generation strategy.

        Returns:
            A generation strategy

        Raises:
            ConfigurationException: if an unknown algorithm was requested
        """
        if config.configuration.algorithm in cls._strategies:
            strategy = cls._strategies.get(config.configuration.algorithm)
            assert strategy, "Strategy cannot be defined as None"
            cls._logger.info("Using strategy: %s", config.configuration.algorithm)
            return strategy()
        raise ConfigurationException("No suitable generation strategy found.")

    @classmethod
    def _get_selection_function(cls) -> SelectionFunction[tsc.TestSuiteChromosome]:
        """Provides a selection function for the selected algorithm.

        Returns:
            A selection function

        Raises:
            ConfigurationException: if an unknown function was requested
        """
        if config.configuration.search_algorithm.selection in cls._selections:
            strategy = cls._selections.get(
                config.configuration.search_algorithm.selection
            )
            assert strategy, "Selection function cannot be defined as None"
            cls._logger.info(
                "Using selection function: %s",
                config.configuration.search_algorithm.selection,
            )
            return strategy()
        raise ConfigurationException("No suitable selection function found.")

    def _get_crossover_function(self) -> CrossOverFunction[tsc.TestSuiteChromosome]:
        """Provides a crossover function for the selected algorithm.

        Returns:
            A crossover function
        """
        self._logger.info("Using crossover function: SinglePointRelativeCrossOver")
        return SinglePointRelativeCrossOver()

    def _get_archive(self, strategy: TestGenerationStrategy) -> arch.Archive:
        if config.configuration.algorithm == config.Algorithm.MIO:
            self._logger.info("Using MIOArchive")
            size = config.configuration.mio.initial_config.number_of_tests_per_target
            return arch.MIOArchive(
                strategy.test_case_fitness_functions,
                initial_size=size,
            )
        # Use CoverageArchive as default, even if it the algorithm does not use it.
        self._logger.info("Using CoverageArchive")
        if config.configuration.algorithm == config.Algorithm.DYNAMOSA:
            # DynaMOSA gradually adds its fitness functions, so we initialize
            # with an empty set.
            return arch.CoverageArchive(OrderedSet())
        return arch.CoverageArchive(OrderedSet(strategy.test_case_fitness_functions))

    def _get_ranking_function(self) -> RankingFunction:
        self._logger.info("Using ranking function: RankBasedPreferenceSorting")
        return RankBasedPreferenceSorting()

    def _get_test_case_fitness_functions(
        self, strategy: TestGenerationStrategy
    ) -> OrderedSet[ff.TestCaseFitnessFunction]:
        """Creates the fitness functions for test cases.

        Args:
            strategy: The currently configured strategy

        Returns:
            A list of fitness functions
        """
        if config.configuration.algorithm in (
            config.Algorithm.DYNAMOSA,
            config.Algorithm.MIO,
            config.Algorithm.MOSA,
            config.Algorithm.RANDOM_TEST_CASE_SEARCH,
            config.Algorithm.WHOLE_SUITE,
        ):
            fitness_functions = bg.create_branch_coverage_fitness_functions(
                self._executor, strategy.branch_goal_pool
            )
            self._logger.info(
                "Instantiated %d fitness functions", len(fitness_functions)
            )
            return fitness_functions
        return OrderedSet()

    def _get_test_suite_fitness_functions(
        self,
    ) -> OrderedSet[ff.TestSuiteFitnessFunction]:
        return OrderedSet([ff.BranchDistanceTestSuiteFitnessFunction(self._executor)])

    def _get_test_suite_coverage_functions(
        self,
    ) -> OrderedSet[ff.TestSuiteCoverageFunction]:
        return OrderedSet([ff.TestSuiteBranchCoverageFunction(self._executor)])

    def _get_test_cluster(self, strategy: TestGenerationStrategy):
        search_alg = config.configuration.search_algorithm
        if search_alg.filter_covered_targets_from_test_cluster:
            # Wrap test cluster in filter.
            return FilteredTestCluster(
                self._test_cluster,
                strategy.archive,
                self._executor.tracer.get_known_data(),
                strategy.test_case_fitness_functions,
            )
        return self._test_cluster

    @staticmethod
    def _get_test_factory(strategy: TestGenerationStrategy):
        return tf.TestFactory(strategy.test_cluster)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a factory to create test case chromosomes."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ordered_set import OrderedSet

import pynguin.configuration as config
import pynguin.ga.chromosomefactory as cf
import pynguin.ga.testcasechromosome as tcc
from pynguin.utils import randomness

if TYPE_CHECKING:
    import pynguin.ga.computations as ff
    import pynguin.ga.testcasefactory as tcf
    import pynguin.generation.algorithms.archive as arch
    import pynguin.testcase.testfactory as tf


class TestCaseChromosomeFactory(
    cf.ChromosomeFactory[tcc.TestCaseChromosome]
):  # pylint:disable=too-few-public-methods.
    """A factory that creates test case chromosomes using the given test case factory.
    Also add the given fitness functions to the newly created test case chromosome."""

    def __init__(
        self,
        test_factory: tf.TestFactory,
        test_case_factory: tcf.TestCaseFactory,
        fitness_functions: OrderedSet[ff.TestCaseFitnessFunction],
    ) -> None:
        """Instantiates a new factory to create test case chromosomes.

        Args:
            test_factory: The internal factory required for the mutation.
            test_case_factory: The internal test case factory required for creation
                               of test cases.
            fitness_functions: The fitness functions that will be added to every
                               newly generated chromosome.
        """
        self._test_factory = test_factory
        self._test_case_factory = test_case_factory
        self._fitness_functions = fitness_functions

    def get_chromosome(self) -> tcc.TestCaseChromosome:
        test_case = self._test_case_factory.get_test_case()
        chrom = tcc.TestCaseChromosome(
            test_case=test_case, test_factory=self._test_factory
        )
        for func in self._fitness_functions:
            chrom.add_fitness_function(func)
        return chrom


class ArchiveReuseTestCaseChromosomeFactory(
    cf.ChromosomeFactory[tcc.TestCaseChromosome]
):  # pylint:disable=too-few-public-methods.
    """Provides test case chromosomes from an archive with some probability,
    otherwise delegates to wrapped chromosome factory."""

    def __init__(
        self,
        delegate: cf.ChromosomeFactory[tcc.TestCaseChromosome],
        archive: arch.Archive,
    ):
        self._delegate = delegate
        self._archive = archive

    def get_chromosome(self) -> tcc.TestCaseChromosome:
        pick_from = self._archive.solutions
        if (
            len(pick_from) > 0
            and randomness.next_float()
            <= config.configuration.seeding.seed_from_archive_probability
        ):
            selected = randomness.choice(pick_from).clone()
            for _ in range(config.configuration.seeding.seed_from_archive_mutations):
                selected.mutate()
            return selected
        return self._delegate.get_chromosome()



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Defines an abstract test case visitor."""

from abc import ABC, abstractmethod


# pylint: disable=too-few-public-methods
class TestCaseVisitor(ABC):
    """An abstract test case visitor."""

    @abstractmethod
    def visit_default_test_case(self, test_case) -> None:
        """Visit a default test case.

        Args:
            test_case: The test case
        """



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a naming scope."""
from __future__ import annotations

from abc import abstractmethod
from collections import defaultdict
from typing import Any, Callable

import pynguin.testcase.variablereference as vr


class AbstractNamingScope:
    """Provides names for objects."""

    @abstractmethod
    def get_name(self, obj) -> str:
        """Get the name for the given object within this scope.

        Args:
            obj: the object for which a name is requested

        Returns:
            the variable name
        """

    @abstractmethod
    def is_known_name(self, obj) -> bool:
        """Does the given object have an assigned name in this scope.

        Args:
            obj: The object to check

        Returns:
            True, iff the object has a name.
        """

    @abstractmethod
    def __len__(self):
        """Return the amount of assigned names."""

    @abstractmethod
    def __iter__(self):
        """Iterate the objects and the associated names"""


class NamingScope(AbstractNamingScope):
    """Maps objects to unique, human friendly names."""

    def __init__(
        self,
        prefix: str = "var",
        new_name_callback: Callable[[Any, str], None] | None = None,
    ) -> None:
        """Initialises the scope.

        Args:
            prefix: The prefix that will be used in all assigned names.
            new_name_callback: Called when a new object is named.
        """
        self._known_names: dict[Any, str] = {}
        self._prefix = prefix
        self._new_name_callback = new_name_callback

    def get_name(self, obj: Any) -> str:
        if obj in self._known_names:
            return self._known_names[obj]

        index = len(self._known_names)
        self._known_names[obj] = name = f"{self._prefix}_{index}"
        if self._new_name_callback is not None:
            self._new_name_callback(obj, name)
        return name

    def __len__(self):
        return len(self._known_names)

    def __iter__(self):
        for obj in self._known_names:
            yield obj, self.get_name(obj)

    def is_known_name(self, obj) -> bool:
        return obj in self._known_names


class VariableTypeNamingScope(AbstractNamingScope):
    """Names variables according to their type."""

    def __init__(self, prefix: str = "var"):
        self._known_variable_names: dict[vr.VariableReference, str] = {}
        self._type_counter: dict[str, int] = defaultdict(int)
        self._prefix = prefix

    def get_name(self, obj: vr.VariableReference) -> str:
        if (name := self._known_variable_names.get(obj)) is not None:
            return name
        type_ = obj.type
        tp_name = self._prefix
        if type_ is not None:
            if isinstance(type_, type):
                # Regular type
                tp_name = snake_case(type_.__name__)
            elif (name_ := getattr(type_, "_name", None)) is not None:
                # Some type hint. Not sure if all have "_name"
                tp_name = snake_case(name_)

        name = f"{tp_name}_{self._type_counter[tp_name]}"
        self._type_counter[tp_name] += 1
        self._known_variable_names[obj] = name
        return name

    def __len__(self):
        return len(self._known_variable_names)

    def __iter__(self):
        yield from self._known_variable_names.items()

    def is_known_name(self, obj) -> bool:
        return obj in self._known_variable_names


def snake_case(name: str) -> str:
    """We assume that we only have to lowercase the first char.

    Args:
        name: the string to camel case

    Returns:
        The cheaply camel cased string.
    """
    assert len(name) > 0, "Cannot snake_case empty string"
    return "".join(["_" + i.lower() if i.isupper() else i for i in name]).lstrip("_")



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""A generic exporter that selects its export strategy based on configuration."""
from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import pynguin.configuration as config
from pynguin.generation.export.noneexporter import NoneExporter
from pynguin.generation.export.pytestexporter import PyTestExporter

if TYPE_CHECKING:
    from pynguin.generation.export.abstractexporter import AbstractTestExporter


# pylint: disable=too-few-public-methods
class ExportProvider:
    """Provides the possibility to export generated tests using a configured strategy"""

    _strategies: dict[config.ExportStrategy, Callable[[bool], AbstractTestExporter]] = {
        config.ExportStrategy.PY_TEST: PyTestExporter,
        config.ExportStrategy.NONE: NoneExporter,
    }

    @classmethod
    def get_exporter(cls, wrap_code: bool = False) -> AbstractTestExporter:
        """Provides an instance of the configured test exporter.

        The flag `wrap_code` indicates whether or not the exported code should be
        wrapped with a `try`-`except`-block.

        Args:
            wrap_code: Whether or not to wrap the generated code

        Returns:
            A test-exporter instance

        Raises:
            Exception: If no appropriate strategy could be found
        """
        strategy = config.configuration.test_case_output.export_strategy
        if strategy in cls._strategies:
            exp = cls._strategies.get(strategy)
            assert exp, "Export strategy cannot be defined as None"
            return exp(wrap_code)
        raise Exception("Unknown export strategy")



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Implements simple constant seeding strategies."""
from __future__ import annotations

import ast
import logging
import os
from abc import abstractmethod
from pathlib import Path
from pkgutil import iter_modules
from typing import Any, Union, cast

from _py_abc import ABCMeta
from ordered_set import OrderedSet
from setuptools import find_packages

from pynguin.utils import randomness

Types = Union[float, int, str]


class _ConstantSeeding(metaclass=ABCMeta):
    """An abstract base class for constant seeding strategies."""

    _logger = logging.getLogger(__name__)

    @property
    def has_strings(self) -> bool:
        """Whether or not we have some strings collected.

        Returns:
            Whether or not we have some strings collected
        """
        return self.has_constants(str)

    @property
    def has_ints(self) -> bool:
        """Whether or not we have some ints collected.

        Returns:
            Whether or not we have some ints collected
        """
        return self.has_constants(int)

    @property
    def has_floats(self) -> bool:
        """Whether or not we have some floats collected.

        Returns:
            Whether or not we have some floats collected
        """
        return self.has_constants(float)

    @abstractmethod
    def has_constants(self, type_: type[Types]) -> bool:
        """Returns whether a constant of a given type exists in the pool.

        Args:
            type_: The type of the constant

        Returns:
            Whether or not a constant of the given type exists  # noqa: DAR202
        """

    @property
    def random_string(self) -> str:
        """Provides a random string from the set of collected strings.

        Returns:
            A random string
        """
        return cast(str, self.random_element(str))

    @property
    def random_int(self) -> int:
        """Provides a random int from the set of collected ints.

        Returns:
            A random int
        """
        return cast(int, self.random_element(int))

    @property
    def random_float(self) -> float:
        """Provides a random float from the set of collected floats.

        Returns:
            A random float
        """
        return cast(float, self.random_element(float))

    @abstractmethod
    def random_element(self, type_: type[Types]) -> Types:
        """Provides a random element of the given type

        Args:
            type_: The given type

        Returns:
            A random element of the given type
        """


class _StaticConstantSeeding(_ConstantSeeding):
    """A simple static constant seeding strategy.

    Extracts all constants from a set of modules by using an AST visitor.
    """

    def __init__(self) -> None:
        self._constants: dict[type[Types], OrderedSet[Types]] = {
            int: OrderedSet(),
            float: OrderedSet(),
            str: OrderedSet(),
        }

    @staticmethod
    def _find_modules(project_path: str | os.PathLike) -> OrderedSet[str]:
        modules: OrderedSet[str] = OrderedSet()
        for package in find_packages(
            project_path,
            exclude=[
                "*.tests",
                "*.tests.*",
                "tests.*",
                "tests",
                "test",
                "test.*",
                "*.test.*",
                "*.test",
            ],
        ):
            package_name = package.replace(".", "/")
            pkg_path = f"{project_path}/{package_name}"
            for info in iter_modules([pkg_path]):
                if not info.ispkg:
                    name = info.name.replace(".", "/")
                    module = f"{package_name}/{name}.py"
                    module_path = Path(project_path) / Path(module)
                    if module_path.exists() and module_path.is_file():
                        # Consider only Python files for constant seeding, as the
                        # seeding relies on the availability of an AST.
                        modules.add(module)
        return modules

    def collect_constants(
        self, project_path: str | os.PathLike
    ) -> dict[type[Types], OrderedSet[Types]]:
        """Collect all constants for a given project.

        Args:
            project_path: The path to the project's root

        Returns:
            A dict of type to set of constants
        """
        assert self._constants is not None
        collector = _ConstantCollector()
        for module in self._find_modules(project_path):
            with open(
                os.path.join(project_path, module), encoding="utf-8"
            ) as module_file:
                try:
                    tree = ast.parse(module_file.read())
                    collector.visit(tree)
                except BaseException as exception:  # pylint: disable=broad-except
                    self._logger.exception("Cannot collect constants: %s", exception)
        self._constants = collector.constants
        return self._constants

    def has_constants(self, type_: type[Types]) -> bool:
        assert self._constants is not None
        return len(self._constants[type_]) > 0

    def random_element(self, type_: type[Types]) -> Types:
        assert self._constants is not None
        return randomness.choice(tuple(self._constants[type_]))


# pylint: disable=invalid-name, missing-function-docstring
class _ConstantCollector(ast.NodeVisitor):
    def __init__(self) -> None:
        self._constants: dict[type[Types], OrderedSet[Types]] = {
            float: OrderedSet(),
            int: OrderedSet(),
            str: OrderedSet(),
        }
        self._string_expressions: OrderedSet[str] = OrderedSet()

    def visit_Constant(self, node: ast.Constant) -> Any:
        if isinstance(node.value, str):
            self._constants[str].add(node.value)
        elif isinstance(node.value, float):
            self._constants[float].add(node.value)
        elif isinstance(node.value, int):
            self._constants[int].add(node.value)
        return self.generic_visit(node)

    def visit_Module(self, node: ast.Module) -> Any:
        return self._visit_doc_string(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        return self._visit_doc_string(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> Any:
        return self._visit_doc_string(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> Any:
        return self._visit_doc_string(node)

    def _visit_doc_string(self, node: ast.AST) -> Any:
        if docstring := ast.get_docstring(node):
            self._string_expressions.add(docstring)
        return self.generic_visit(node)

    @property
    def constants(self) -> dict[type[Types], OrderedSet[Types]]:
        """Provides the collected constants.

        Returns:
            The collected constants
        """
        self._remove_docstrings()
        return self._constants

    def _remove_docstrings(self) -> None:
        self._constants[str] -= self._string_expressions


class DynamicConstantSeeding(_ConstantSeeding):
    """Provides a dynamic pool and methods to add and retrieve values.

    The methods in this class are added to the module under test during an instruction
    phase before the main algorithm is executed. During this instruction phase,
    bytecode is added to the module under test which executes the methods adding
    values to the dynamic pool. The instrumentation is implemented in the module
    dynamicseedinginstrumentation.py.

    During the test generation process when a new value of one of the supported types
    is needed, this module provides methods to get values from the dynamic pool
    instead of randomly generating a new one.
    """

    _string_functions_lookup = {
        "isalnum": lambda value: f"{value}!" if value.isalnum() else "isalnum",
        "islower": lambda value: value.upper() if value.islower() else value.lower(),
        "isupper": lambda value: value.lower() if value.isupper() else value.upper(),
        "isdecimal": lambda value: "non_decimal" if value.isdecimal() else "0123456789",
        "isalpha": lambda value: f"{value}1" if value.isalpha() else "isalpha",
        "isdigit": lambda value: f"{value}_" if value.isdigit() else "0",
        "isidentifier": lambda value: f"{value}!"
        if value.isidentifier()
        else "is_Identifier",
        "isnumeric": lambda value: f"{value}A" if value.isnumeric() else "012345",
        "isprintable": lambda value: f"{value}{os.linesep}"
        if value.isprintable()
        else "is_printable",
        "isspace": lambda value: f"{value}a" if value.isspace() else "   ",
        "istitle": lambda value: f"{value} AAA" if value.istitle() else "Is Title",
    }

    def __init__(self):
        self._dynamic_pool: dict[type[Types], OrderedSet[Types]] = {
            int: OrderedSet(),
            float: OrderedSet(),
            str: OrderedSet(),
        }

    def reset(self) -> None:
        """Delete all currently stored dynamic constants"""
        for elem in self._dynamic_pool.values():
            elem.clear()

    def has_constants(self, type_: type[Types]) -> bool:
        assert type_ in self._dynamic_pool
        return len(self._dynamic_pool[type_]) > 0

    def random_element(self, type_: type[Types]) -> Types:
        return randomness.choice(tuple(self._dynamic_pool[type_]))

    def add_value(self, value: Types):
        """Adds the given value to the corresponding set of the dynamic pool.

        Args:
            value: The value to add.
        """
        if isinstance(
            value, bool
        ):  # needed because True and False are accepted as ints otherwise
            return
        if type(value) in self._dynamic_pool:
            self._dynamic_pool[type(value)].add(value)

    def add_value_for_strings(self, value: str, name: str):
        """Add a value of a string.

        Args:
            value: The value
            name: The string
        """
        if not isinstance(value, str):
            return
        self._dynamic_pool[str].add(value)
        self._dynamic_pool[str].add(self._string_functions_lookup[name](value))


static_constant_seeding = _StaticConstantSeeding()
dynamic_constant_seeding = DynamicConstantSeeding()



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a no op exporter."""
from __future__ import annotations

import os
from typing import TYPE_CHECKING

from pynguin.generation.export.abstractexporter import AbstractTestExporter

if TYPE_CHECKING:
    from pynguin.testcase import testcase as tc


# pylint: disable=too-few-public-methods
class NoneExporter(AbstractTestExporter):
    """An exporter, which does basically nothing."""

    def export_sequences(self, path: str | os.PathLike, test_cases: list[tc.TestCase]):
        pass



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Implements seeding of the initial population with previously existing testcases."""
from __future__ import annotations

import ast
import logging
import os
from typing import TYPE_CHECKING, AnyStr

import pynguin.analyses.seeding.testimport.ast_to_statement as ats
import pynguin.configuration as config
import pynguin.ga.testcasechromosome as tcc
import pynguin.testcase.defaulttestcase as dtc
import pynguin.testcase.testfactory as tf
import pynguin.utils.statistics.statistics as stat
from pynguin.utils import randomness
from pynguin.utils.statistics.runtimevariable import RuntimeVariable

if TYPE_CHECKING:
    from pynguin.setup.testcluster import TestCluster


class _InitialPopulationSeeding:
    """Class for seeding the initial population with previously existing testcases."""

    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self._testcases: list[dtc.DefaultTestCase] = []
        self._test_cluster: TestCluster

    @property
    def test_cluster(self) -> TestCluster:
        """Provides the test cluster.

        Returns:
            The test cluster
        """
        return self._test_cluster

    @test_cluster.setter
    def test_cluster(self, test_cluster: TestCluster):
        self._test_cluster = test_cluster

    def get_ast_tree(
        self, module_path: AnyStr | "os.PathLike[AnyStr]"
    ) -> ast.Module | None:

        """Returns the ast tree from a module

        Args:
            module_path: The path to the project's root

        Returns:
            The ast tree of the given module.
        """
        module_name = config.configuration.module_name.rsplit(".", maxsplit=1)[-1]
        self._logger.debug("Module name: %s", module_name)
        result: list[AnyStr] = []
        for root, _, files in os.walk(module_path):
            for name in files:
                assert isinstance(name, str)
                if module_name in name and "test_" in name:
                    result.append(os.path.join(root, name))
                    break
        try:
            if len(result) > 0:
                self._logger.debug("Module name found: %s", result[0])
                stat.track_output_variable(RuntimeVariable.SuitableTestModule, True)
                with open(result[0], encoding="utf-8") as module_file:
                    return ast.parse(module_file.read())
            else:
                self._logger.debug("No suitable test module found.")
                stat.track_output_variable(RuntimeVariable.SuitableTestModule, False)
                return None
        except BaseException as exception:  # pylint: disable=broad-except
            self._logger.exception("Cannot read module: %s", exception)
            stat.track_output_variable(RuntimeVariable.SuitableTestModule, False)
            return None

    def collect_testcases(self, module_path: AnyStr | "os.PathLike[AnyStr]") -> None:
        """Collect all test cases from a module.

        Args:
            module_path: Path to the module to collect the test cases from
        """
        tree = self.get_ast_tree(module_path)
        if tree is None:
            config.configuration.seeding.initial_population_seeding = False
            self._logger.info("Provided testcases are not used.")
            return
        transformer = ats.AstToTestCaseTransformer(
            self._test_cluster,
            config.configuration.test_case_output.assertion_generation
            != config.AssertionGenerator.NONE,
        )
        transformer.visit(tree)
        self._testcases = transformer.testcases
        stat.track_output_variable(RuntimeVariable.FoundTestCases, len(self._testcases))
        if not self._testcases:
            config.configuration.seeding.initial_population_seeding = False
            self._logger.info("None of the provided test cases can be parsed.")
        else:
            self._logger.info(
                "Number successfully collected test cases: %s", len(self._testcases)
            )
        stat.track_output_variable(
            RuntimeVariable.CollectedTestCases, len(self._testcases)
        )
        self._mutate_testcases_initially()

    def _mutate_testcases_initially(self):
        """Mutates the initial population."""
        test_factory = tf.TestFactory(self.test_cluster)
        for _ in range(0, config.configuration.seeding.initial_population_mutations):
            for testcase in self._testcases:
                testcase_wrapper = tcc.TestCaseChromosome(testcase, test_factory)
                testcase_wrapper.mutate()
                if not testcase_wrapper.test_case.statements:
                    self._testcases.remove(testcase)

    @property
    def seeded_testcase(self) -> dtc.DefaultTestCase:
        """Provides a random seeded test case.

        Returns:
            A random test case
        """
        return self._testcases[randomness.next_int(0, len(self._testcases))]

    @property
    def has_tests(self) -> bool:
        """Whether or not test cases have been found.

        Returns:
            Whether or not test cases have been found
        """
        return len(self._testcases) > 0


initialpopulationseeding = _InitialPopulationSeeding()



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""
Provides classes for runtime instrumentation.
Inspired by https://github.com/agronholm/typeguard/blob/master/typeguard/importhook.py
"""
from __future__ import annotations

import logging
import sys
from importlib.abc import FileLoader, MetaPathFinder
from importlib.machinery import ModuleSpec, SourceFileLoader
from inspect import isclass
from types import CodeType
from typing import TYPE_CHECKING, cast

import pynguin.configuration as config
from pynguin.analyses.seeding.constantseeding import dynamic_constant_seeding
from pynguin.instrumentation.instrumentation import (
    BranchCoverageInstrumentation,
    DynamicSeedingInstrumentation,
)

if TYPE_CHECKING:
    from pynguin.instrumentation.instrumentation import Instrumentation
    from pynguin.testcase.execution import ExecutionTracer


class InstrumentationLoader(SourceFileLoader):
    """A loader that instruments the module after execution."""

    def __init__(self, fullname, path, tracer: ExecutionTracer):
        super().__init__(fullname, path)
        self._tracer = tracer

    def exec_module(self, module):
        self._tracer.reset()
        super().exec_module(module)
        self._tracer.store_import_trace()

    def get_code(self, fullname) -> CodeType:
        """Add instrumentation instructions to the code of the module
        before it is executed.

        Args:
            fullname: The name of the module

        Returns:
            The modules code blocks
        """
        to_instrument = cast(CodeType, super().get_code(fullname))
        assert to_instrument, "Failed to get code object of module."
        instrumentations: list[Instrumentation] = [
            BranchCoverageInstrumentation(self._tracer)
        ]
        if config.configuration.seeding.dynamic_constant_seeding:
            instrumentations.append(
                DynamicSeedingInstrumentation(dynamic_constant_seeding)
            )

        for instrumentation in instrumentations:
            to_instrument = instrumentation.instrument_module(to_instrument)
        return to_instrument


class InstrumentationFinder(MetaPathFinder):
    """
    A meta path finder which wraps another pathfinder.
    It receives all import requests and intercepts the ones for the modules that
    should be instrumented.
    """

    _logger = logging.getLogger(__name__)

    def __init__(
        self, original_pathfinder, module_to_instrument: str, tracer: ExecutionTracer
    ) -> None:
        """Wraps the given path finder.

        Args:
            original_pathfinder: the original pathfinder that is wrapped.
            module_to_instrument: the name of the module, that should be instrumented.
            tracer: the execution tracer
        """
        self._module_to_instrument = module_to_instrument
        self._original_pathfinder = original_pathfinder
        self._tracer = tracer

    def _should_instrument(self, module_name: str):
        return module_name == self._module_to_instrument

    def find_spec(self, fullname: str, path=None, target=None):
        """Try to find a spec for the given module.

        If the original path finder accepts the request, we take the spec and replace
        the loader.

        Args:
            fullname: The full name of the module
            path: The path
            target: The target

        Returns:
            An optional ModuleSpec
        """
        if self._should_instrument(fullname):
            spec: ModuleSpec = self._original_pathfinder.find_spec(
                fullname, path, target
            )
            if spec is not None:
                if isinstance(spec.loader, FileLoader):
                    spec.loader = InstrumentationLoader(
                        spec.loader.name, spec.loader.path, self._tracer
                    )
                    return spec
                self._logger.error(
                    "Loader for module under test is not a FileLoader,"
                    " cannot instrument."
                )

        return None


class ImportHookContextManager:
    """A simple context manager for using the import hook."""

    def __init__(self, hook: MetaPathFinder):
        self.hook = hook

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.uninstall()

    def uninstall(self):
        """Remove the installed hook."""
        try:
            sys.meta_path.remove(self.hook)
        except ValueError:
            pass  # already removed


def install_import_hook(
    module_to_instrument: str, tracer: ExecutionTracer
) -> ImportHookContextManager:
    """Install the InstrumentationFinder in the meta path.

    Args:
        module_to_instrument: The module that shall be instrumented.
        tracer: The tracer where the instrumentation should report its data.

    Returns:
        a context manager which can be used to uninstall the hook.

    Raises:
        RuntimeError: In case a PathFinder could not be found
    """
    to_wrap = None
    for finder in sys.meta_path:
        if (
            isclass(finder)
            and finder.__name__ == "PathFinder"
            and hasattr(finder, "find_spec")
        ):
            to_wrap = finder
            break

    if not to_wrap:
        raise RuntimeError("Cannot find a PathFinder in sys.meta_path")

    hook = InstrumentationFinder(to_wrap, module_to_instrument, tracer)
    sys.meta_path.insert(0, hook)
    return ImportHookContextManager(hook)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides tracking of statistics for various variables and types."""
from __future__ import annotations

import logging
import queue
import time
from typing import Any, Generator

import pynguin.configuration as config
import pynguin.ga.chromosome as chrom
import pynguin.utils.statistics.outputvariablefactory as ovf
import pynguin.utils.statistics.statisticsbackend as sb
from pynguin.utils.statistics.runtimevariable import RuntimeVariable


class _StatisticsTracker:
    """A singleton tracker for statistics."""

    def __init__(self) -> None:
        self._variables: queue.Queue = queue.Queue()
        self._search_statistics: _SearchStatistics = _SearchStatistics()

    def reset(self) -> None:
        """Reset the tracker (necessary for testing only)."""
        self._variables = queue.Queue()
        self._search_statistics = _SearchStatistics()

    def track_output_variable(self, runtime_variable: RuntimeVariable, value: Any):
        """Tracks a run-time variable for output.

        Args:
            runtime_variable: The run-time variable
            value: The value to track for the variable
        """
        self._variables.put((runtime_variable, value))

    @property
    def variables(self) -> queue.Queue:
        """Provides the queue of tracked variables.

        Returns:
            The queue of tracked variables
        """
        return self._variables

    @property
    def variables_generator(self) -> Generator[tuple[RuntimeVariable, Any], None, None]:
        """Provides a generator.

        Yields:
            A generator for iteration
        """
        while not self._variables.empty():
            yield self._variables.get()

    @property
    def search_statistics(self) -> _SearchStatistics:
        """Provides the internal search statistics instance.

        Returns:
            The search statistics instance
        """
        return self._search_statistics

    def set_sequence_start_time(self, start_time: int) -> None:
        """This should only be called once, before any sequence data was generated.

        Args:
            start_time: the start time
        """
        self._search_statistics.set_sequence_output_variable_start_time(start_time)

    def current_individual(self, individual: chrom.Chromosome) -> None:
        """Called when a new individual is sent.

        The individual represents the best individual of the current generation.

        Args:
            individual: The best individual of the current generation
        """
        self._search_statistics.current_individual(individual)

    def set_output_variable(self, variable: sb.OutputVariable) -> None:
        """Sets an output variable to a value directly

        Args:
            variable: The variable to be set
        """
        self._search_statistics.set_output_variable(variable)

    def update_output_variable(self, variable: sb.OutputVariable) -> None:
        """Updates an output variable with a value.

        Args:
            variable: The variable to update
        """
        self._search_statistics.update_output_variable(variable)

    def set_output_variable_for_runtime_variable(
        self, variable: RuntimeVariable, value: Any
    ) -> None:
        """Sets an output variable to a value directly

        Args:
            variable: The variable to be set
            value: the value to be set
        """
        self._search_statistics.set_output_variable_for_runtime_variable(
            variable, value
        )

    def update_output_variable_for_runtime_variable(
        self, variable: RuntimeVariable, value: Any
    ) -> None:
        """Updates an output variable with a value directly

        Args:
            variable: The variable to update
            value: The value to add
        """
        self._search_statistics.update_output_variable_for_runtime_variable(
            variable, value
        )

    @property
    def output_variables(self) -> dict[str, sb.OutputVariable]:
        """Provides the output variables.

        Returns:
            The output variables
        """
        return self._search_statistics.output_variables

    def write_statistics(self) -> bool:
        """Write result to disk using selected backend

        Returns:
            True if the writing was successful
        """
        return self._search_statistics.write_statistics()


class _SearchStatistics:
    """A singleton of SearchStatistics collects all the data values reported.

    Because we cannot guarantee a singleton here without making the code too crazy,
    the only instance of this class that shall exist throughout the whole framework
    is in the `StatisticsTracker`.  The `StatisticsTracker` provides public methods
    for all public methods of this class, which delegate to its instance.
    """

    _logger = logging.getLogger(__name__)

    def __init__(self):
        self._backend: None | (
            sb.AbstractStatisticsBackend
        ) = self._initialise_backend()
        self._output_variables: dict[str, sb.OutputVariable] = {}
        self._variable_factories: dict[str, ovf.ChromosomeOutputVariableFactory] = {}
        self._sequence_output_variable_factories: dict[
            str, ovf.SequenceOutputVariableFactory
        ] = {}
        self._init_factories()
        self.set_output_variable_for_runtime_variable(
            RuntimeVariable.RandomSeed, config.configuration.seeding.seed
        )
        self._fill_sequence_output_variable_factories()
        self._start_time = time.time_ns()
        self.set_sequence_output_variable_start_time(self._start_time)
        self._best_individual: chrom.Chromosome | None = None

    @staticmethod
    def _initialise_backend() -> sb.AbstractStatisticsBackend | None:
        backend = config.configuration.statistics_output.statistics_backend
        if backend == config.StatisticsBackend.CONSOLE:
            return sb.ConsoleStatisticsBackend()
        if backend == config.StatisticsBackend.CSV:
            return sb.CSVStatisticsBackend()
        return None

    def _init_factories(self) -> None:
        self._variable_factories[
            RuntimeVariable.Length.name
        ] = self._ChromosomeLengthOutputVariableFactory()
        self._variable_factories[
            RuntimeVariable.Size.name
        ] = self._ChromosomeSizeOutputVariableFactory()
        self._variable_factories[
            RuntimeVariable.Coverage.name
        ] = self._ChromosomeCoverageOutputVariableFactory()
        self._variable_factories[
            RuntimeVariable.Fitness.name
        ] = self._ChromosomeFitnessOutputVariableFactory()

    def _fill_sequence_output_variable_factories(self) -> None:
        self._sequence_output_variable_factories[
            RuntimeVariable.CoverageTimeline.name
        ] = self._CoverageSequenceOutputVariableFactory()
        self._sequence_output_variable_factories[
            RuntimeVariable.SizeTimeline.name
        ] = self._SizeSequenceOutputVariableFactory()
        self._sequence_output_variable_factories[
            RuntimeVariable.LengthTimeline.name
        ] = self._LengthSequenceOutputVariableFactory()
        self._sequence_output_variable_factories[
            RuntimeVariable.FitnessTimeline.name
        ] = self._FitnessSequenceOutputVariableFactory()
        self._sequence_output_variable_factories[
            RuntimeVariable.TotalExceptionsTimeline.name
        ] = ovf.DirectSequenceOutputVariableFactory.get_integer(
            RuntimeVariable.TotalExceptionsTimeline
        )

    def set_sequence_output_variable_start_time(self, start_time: int) -> None:
        """Set start time for sequence data.

        Args:
            start_time: the start time
        """
        for factory in self._sequence_output_variable_factories.values():
            factory.set_start_time(start_time)

    def current_individual(self, individual: chrom.Chromosome) -> None:
        """Called when a new individual is sent.

        The individual represents the best individual of the current generation.

        Args:
            individual: The best individual of the current generation
        """
        if not self._backend:
            return

        if not isinstance(individual, chrom.Chromosome):
            self._logger.warning("SearchStatistics expected a TestSuiteChromosome")
            return

        self._logger.debug("Received individual")
        self._best_individual = individual
        for variable_factory in self._variable_factories.values():
            self.set_output_variable(variable_factory.get_variable(individual))
        for seq_variable_factory in self._sequence_output_variable_factories.values():
            seq_variable_factory.update(individual)

    def set_output_variable(self, variable: sb.OutputVariable) -> None:
        """Sets an output variable to a value directly

        Args:
            variable: The variable to be set
        """
        if variable.name in self._sequence_output_variable_factories:
            var = self._sequence_output_variable_factories[variable.name]
            assert isinstance(var, ovf.DirectSequenceOutputVariableFactory)
            var.set_value(variable.value)
        else:
            self._output_variables[variable.name] = variable

    def update_output_variable(self, variable: sb.OutputVariable) -> None:
        """Updates an output variable with a new value

        Args:
            variable: The variable to update
        """
        if variable.name not in self._sequence_output_variable_factories:
            raise AssertionError("Can only be called on sequence variable.")
        var = self._sequence_output_variable_factories[variable.name]
        assert isinstance(var, ovf.DirectSequenceOutputVariableFactory)
        var.update_value(variable.value)

    def set_output_variable_for_runtime_variable(
        self, variable: RuntimeVariable, value: Any
    ) -> None:
        """Sets an output variable to a value directly

        Args:
            variable: The variable to be set
            value: the value to be set
        """
        self.set_output_variable(sb.OutputVariable(name=variable.name, value=value))

    def update_output_variable_for_runtime_variable(
        self, variable: RuntimeVariable, value: Any
    ) -> None:
        """Updates an output variable with a new value

        Args:
            variable: The variable to update
            value: The value to add
        """
        self.update_output_variable(sb.OutputVariable(name=variable.name, value=value))

    @property
    def output_variables(self) -> dict[str, sb.OutputVariable]:
        """Provides the output variables.

        Returns:
            The output variables
        """
        return self._output_variables

    def _get_output_variables(
        self, individual, skip_missing: bool = True
    ) -> dict[str, sb.OutputVariable]:
        output_variables_map: dict[str, sb.OutputVariable] = {}

        for variable in config.configuration.statistics_output.output_variables:
            variable_name = variable.name
            if variable_name in self._output_variables:
                # Values directly sent
                output_variables_map[variable_name] = self._output_variables[
                    variable_name
                ]
            elif variable_name in self._variable_factories:
                # Values extracted from the individual
                output_variables_map[variable_name] = self._variable_factories[
                    variable_name
                ].get_variable(individual)
            elif variable_name in self._sequence_output_variable_factories:
                # Time related values, which will be expanded in a list of values
                # through time
                for var in self._sequence_output_variable_factories[
                    variable_name
                ].get_output_variables():
                    output_variables_map[var.name] = var
            elif skip_missing:
                # if variable does not exist, return an empty value instead
                output_variables_map[variable_name] = sb.OutputVariable(
                    name=variable_name, value=""
                )
            else:
                self._logger.error(
                    "No obtained value for output variable %s", variable_name
                )
                return {}

        return output_variables_map

    def write_statistics(self) -> bool:
        """Write result to disk using selected backend

        Returns:
            True if the writing was successful
        """
        self._logger.info("Writing statistics")
        # reinitialise backend to be sure we got the correct one, prone to failure
        # due to global-object pattern otherwise.
        self._backend = self._initialise_backend()
        if not self._backend:
            return False

        self._output_variables[RuntimeVariable.TotalTime.name] = sb.OutputVariable(
            name=RuntimeVariable.TotalTime.name,
            value=time.time_ns() - self._start_time,
        )

        if not self._best_individual:
            self._logger.error(
                "No statistics has been saved because Pynguin failed to generate any "
                "test case"
            )
            return False

        individual = self._best_individual
        output_variables_map = self._get_output_variables(individual)
        self._backend.write_data(output_variables_map)
        return True

    class _ChromosomeLengthOutputVariableFactory(ovf.ChromosomeOutputVariableFactory):
        def __init__(self) -> None:
            super().__init__(RuntimeVariable.Length)

        def get_data(self, individual: chrom.Chromosome) -> int:
            return individual.length()

    class _ChromosomeSizeOutputVariableFactory(ovf.ChromosomeOutputVariableFactory):
        def __init__(self) -> None:
            super().__init__(RuntimeVariable.Size)

        def get_data(self, individual: chrom.Chromosome) -> int:
            return individual.size()

    class _ChromosomeCoverageOutputVariableFactory(ovf.ChromosomeOutputVariableFactory):
        def __init__(self) -> None:
            super().__init__(RuntimeVariable.Coverage)

        def get_data(self, individual: chrom.Chromosome) -> float:
            return individual.get_coverage()

    class _ChromosomeFitnessOutputVariableFactory(ovf.ChromosomeOutputVariableFactory):
        def __init__(self) -> None:
            super().__init__(RuntimeVariable.Fitness)

        def get_data(self, individual: chrom.Chromosome) -> float:
            return individual.get_fitness()

    class _CoverageSequenceOutputVariableFactory(
        ovf.DirectSequenceOutputVariableFactory
    ):
        def __init__(self) -> None:
            super().__init__(RuntimeVariable.CoverageTimeline, 0.0)

        def get_value(self, individual: chrom.Chromosome) -> float:
            return individual.get_coverage()

    class _SizeSequenceOutputVariableFactory(ovf.DirectSequenceOutputVariableFactory):
        def __init__(self) -> None:
            super().__init__(RuntimeVariable.SizeTimeline, 0)

        def get_value(self, individual: chrom.Chromosome) -> int:
            return individual.size()

    class _LengthSequenceOutputVariableFactory(ovf.DirectSequenceOutputVariableFactory):
        def __init__(self) -> None:
            super().__init__(RuntimeVariable.LengthTimeline, 0)

        def get_value(self, individual: chrom.Chromosome) -> int:
            return individual.length()

    class _FitnessSequenceOutputVariableFactory(
        ovf.DirectSequenceOutputVariableFactory
    ):
        def __init__(self) -> None:
            super().__init__(RuntimeVariable.FitnessTimeline, 0.0)

        def get_value(self, individual: chrom.Chromosome) -> float:
            return individual.get_fitness()


statistics_tracker = _StatisticsTracker()
track_output_variable = statistics_tracker.track_output_variable
variables = statistics_tracker.variables
variables_generator = statistics_tracker.variables_generator
search_statistics = statistics_tracker.search_statistics
set_sequence_start_time = statistics_tracker.set_sequence_start_time
current_individual = statistics_tracker.current_individual
set_output_variable = statistics_tracker.set_output_variable
set_output_variable_for_runtime_variable = (
    statistics_tracker.set_output_variable_for_runtime_variable
)
update_output_variable = search_statistics.update_output_variable
update_output_variable_for_runtime_variable = (
    search_statistics.update_output_variable_for_runtime_variable
)
output_variables = statistics_tracker.output_variables
write_statistics = statistics_tracker.write_statistics
reset = statistics_tracker.reset



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a base implementation of a variable in a test case."""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any

from pynguin.utils import type_utils
from pynguin.utils.type_utils import is_type_unknown

if TYPE_CHECKING:
    import pynguin.testcase.testcase as tc
    import pynguin.utils.generic.genericaccessibleobject as gao
    import pynguin.utils.namingscope as ns


class Reference(metaclass=ABCMeta):
    """Represents something that can be referenced in a test case.

    For example:
        foo_0 = Foo()
        int_0 = 42
        foo_0.bar = int_0

    Here, foo_0, int_0 and foo_0.bar are references.
    """

    def __init__(self, tp_: type | None) -> None:
        self._type = tp_

    @property
    def type(self) -> type | None:
        """Provides the type of this reference.

        Returns:
            The type of this reference
        """
        return self._type

    def is_primitive(self) -> bool:
        """Does this variable reference represent a primitive type.

        Returns:
            True if the variable is a primitive
        """
        return type_utils.is_primitive_type(self._type)

    def is_none_type(self) -> bool:
        """Is this variable reference of type none, i.e. it does not return anything.

        Returns:
            True if this variable is a none type
        """
        return type_utils.is_none_type(self._type)

    def is_type_unknown(self) -> bool:
        """Is the type of this variable unknown?

        Returns:
            True if this variable has unknown type
        """
        return is_type_unknown(self._type)

    @abstractmethod
    def get_names(
        self,
        variable_names: ns.AbstractNamingScope,
        module_names: ns.AbstractNamingScope,
    ) -> list[str]:
        """Get the names involved when addressing this reference.

        Args:
            variable_names: Mapping for variable names.
            module_names: Mapping for modules.

        Returns:
            A list of the elements in the reference names, i.e.,
            resolving a reference that represents "module_0.Foo.bar" gives
            ["module_0", "Foo", "bar"]
        """

    @abstractmethod
    def clone(self, memo: dict[VariableReference, VariableReference]) -> Reference:
        """Clone this reference.

        Args:
            memo: A mapping from variables in this test case to their corresponding
                variable in the new test case.

        Returns:
            A clone of this reference.
        """

    @abstractmethod
    def structural_eq(
        self, other: Any, memo: dict[VariableReference, VariableReference]
    ) -> bool:
        """Compare if this reference is the same as the other.

        Args:
            other: The variable to compare
            memo: A mapping from variables in this test case to their corresponding
                variable in the compared test case.

        Returns:
            True, iff this variable is the same as the other and points to the same
            location.
        """

    @abstractmethod
    def structural_hash(self) -> int:
        """Required for structural_eq to work.

        Returns:
            A hash value.
        """

    @abstractmethod
    def get_variable_reference(self) -> VariableReference | None:
        """Provide the variable reference used in this reference

        Returns: The variable reference used here, if any.
        """

    @abstractmethod
    def replace_variable_reference(
        self, old: VariableReference, new: VariableReference
    ) -> None:
        """Replace the old variable with the new variable.

        Args:
            old: The old variable reference
            new: The new variable reference
        """


class VariableReference(Reference):
    """A reference to a variable declared in a test case.

    For example:
        int_0 = 5

    Note that this class does NOT implement eq/hash because we use object identity
    to check for equality. The other reference do implement eq/hash.
    """

    def __init__(self, test_case: tc.TestCase, tp_: type | None):
        super().__init__(tp_)
        self._test_case = test_case
        self._distance = 0

    @property
    def test_case(self) -> tc.TestCase:
        """Provides the test case in which this variable reference is used.

        Returns:
            The containing test case
        """
        return self._test_case

    @property
    def distance(self) -> int:
        """Distance metric used to select variables for mutation based on how close
        they are to the subject under test.

        Returns:
            The distance value
        """
        return self._distance

    @distance.setter
    def distance(self, distance: int) -> None:
        """Set the distance metric.

        Args:
            distance: The new distance value
        """
        self._distance = distance

    def get_names(
        self,
        variable_names: ns.AbstractNamingScope,
        module_names: ns.AbstractNamingScope,
    ) -> list[str]:
        return [variable_names.get_name(self)]

    def clone(
        self, memo: dict[VariableReference, VariableReference]
    ) -> VariableReference:
        return memo[self]

    def structural_eq(
        self, other: Any, memo: dict[VariableReference, VariableReference]
    ) -> bool:
        if not isinstance(other, VariableReference):
            return False
        return self._type == other._type and memo[self] == other

    def structural_hash(self) -> int:
        return 31 * 17 + hash(self._type)

    def get_statement_position(self) -> int:
        """Provides the position of the statement which defines this variable reference
        in the test case.

        Raises:
            Exception: if the statement is not found in the test case

        Returns:
            The position  # noqa: DAR202
        """
        for idx, stmt in enumerate(self._test_case.statements):
            if stmt.ret_val == self:
                return idx
        raise Exception(
            "Variable reference is not declared in the test case in which it is used"
        )

    def get_variable_reference(self) -> VariableReference | None:
        return self

    def replace_variable_reference(
        self, old: VariableReference, new: VariableReference
    ) -> None:
        # We can't replace ourselves.
        return


class FieldReference(Reference):
    """A reference to a non-static field."""

    def __init__(self, source: Reference, field: gao.GenericField):
        super().__init__(field.generated_type())
        self._source = source
        self._field = field

    @property
    def source(self) -> Reference:
        """Provide the source.

        Returns:
            The source.
        """
        return self._source

    @property
    def field(self) -> gao.GenericField:
        """Provide the field.

        Returns:
            The field
        """
        return self._field

    def get_names(
        self,
        variable_names: ns.AbstractNamingScope,
        module_names: ns.AbstractNamingScope,
    ) -> list[str]:
        lst = self._source.get_names(variable_names, module_names)
        lst.append(self._field.field)
        return lst

    def clone(self, memo: dict[VariableReference, VariableReference]) -> FieldReference:
        return FieldReference(self._source.clone(memo), self._field)

    def structural_eq(
        self, other: Any, memo: dict[VariableReference, VariableReference]
    ) -> bool:
        if not isinstance(other, FieldReference):
            return False
        return self._field == other._field and self._source.structural_eq(
            other._source, memo
        )

    def structural_hash(self) -> int:
        return hash((self._field, self._source.structural_hash()))

    def __eq__(self, other):
        if not isinstance(other, FieldReference):
            return False
        return self._field == other._field and self._source == other._source

    def __hash__(self):
        return hash((self._field, self._source))

    def get_variable_reference(self) -> VariableReference | None:
        return self._source.get_variable_reference()

    def replace_variable_reference(
        self, old: VariableReference, new: VariableReference
    ) -> None:
        if self._source == old:
            self._source = new
        else:
            self._source.replace_variable_reference(old, new)


class StaticFieldReference(Reference):
    """A reference to a static field of a class."""

    def __init__(self, field: gao.GenericStaticField):
        super().__init__(field.generated_type())
        self._field = field

    @property
    def field(self) -> gao.GenericStaticField:
        """Provide the field.

        Returns:
            The field
        """
        return self._field

    def get_names(
        self,
        variable_names: ns.AbstractNamingScope,
        module_names: ns.AbstractNamingScope,
    ) -> list[str]:
        assert self._field.owner is not None
        return [
            module_names.get_name(self._field.owner.__module__),
            self._field.owner.__name__,
            self._field.field,
        ]

    def clone(
        self, memo: dict[VariableReference, VariableReference]
    ) -> StaticFieldReference:
        return StaticFieldReference(self._field)

    def structural_eq(
        self, other: Any, memo: dict[VariableReference, VariableReference]
    ) -> bool:
        if not isinstance(other, StaticFieldReference):
            return False
        return self._field == other._field

    def structural_hash(self) -> int:
        return hash(self._field)

    def __eq__(self, other):
        return self.structural_eq(other, {})

    def __hash__(self):
        return self.structural_hash()

    def get_variable_reference(self) -> VariableReference | None:
        return None

    def replace_variable_reference(
        self, old: VariableReference, new: VariableReference
    ) -> None:
        return


class StaticModuleFieldReference(Reference):
    """A reference to a static module field."""

    # TODO(fk) combine with regular static field?

    def __init__(self, field: gao.GenericStaticModuleField):
        super().__init__(field.generated_type())
        self._field = field

    @property
    def field(self) -> gao.GenericStaticModuleField:
        """Provide the field.

        Returns:
            The field
        """
        return self._field

    def get_names(
        self,
        variable_names: ns.AbstractNamingScope,
        module_names: ns.AbstractNamingScope,
    ) -> list[str]:
        return [module_names.get_name(self._field.module), self._field.field]

    def clone(
        self, memo: dict[VariableReference, VariableReference]
    ) -> StaticModuleFieldReference:
        return StaticModuleFieldReference(self._field)

    def structural_eq(
        self, other: Any, memo: dict[VariableReference, VariableReference]
    ) -> bool:
        if not isinstance(other, StaticModuleFieldReference):
            return False
        return self._field == other._field

    def structural_hash(self) -> int:
        return hash(self._field)

    def __eq__(self, other):
        return self.structural_eq(other, {})

    def __hash__(self):
        return self.structural_hash()

    def get_variable_reference(self) -> VariableReference | None:
        return None

    def replace_variable_reference(
        self, old: VariableReference, new: VariableReference
    ) -> None:
        return



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a base class for assertions."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import pynguin.testcase.variablereference as vr


class Assertion:
    """Base class for assertions."""

    @abstractmethod
    def accept(self, visitor: AssertionVisitor) -> None:
        """Accept an assertion visitor.

        Args:
            visitor: the visitor that is accepted.
        """

    @abstractmethod
    def clone(
        self, memo: dict[vr.VariableReference, vr.VariableReference]
    ) -> Assertion:
        """Clone this assertion.

        Args:
            memo: Mapping from old to new variables.

        Returns: the cloned assertion
        """

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass  # pragma: no cover

    @abstractmethod
    def __hash__(self) -> int:
        pass  # pragma: no cover


class ReferenceAssertion(Assertion, ABC):
    """An assertion on a single reference."""

    def __init__(self, source: vr.Reference):
        self._source = source

    @property
    def source(self) -> vr.Reference:
        """Provides the reference on which we assert something.

        Returns:
            The reference on which we assert something.
        """
        return self._source


class NotNoneAssertion(ReferenceAssertion):
    """An assertion that a reference is not None.

    For example:
        assert var_0 is not None
    """

    def accept(self, visitor: AssertionVisitor) -> None:
        visitor.visit_not_none_assertion(self)

    def clone(
        self, memo: dict[vr.VariableReference, vr.VariableReference]
    ) -> NotNoneAssertion:
        return NotNoneAssertion(self._source.clone(memo))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, NotNoneAssertion) and self._source == other._source

    def __hash__(self) -> int:
        return hash(self._source)


class FloatAssertion(ReferenceAssertion):
    """An assertion on the float value of a reference.

    For example:
        assert float_0 == pytest.approx(42, rel=0.01, abs=0.01)
    """

    def __init__(self, source: vr.Reference, value: float):
        super().__init__(source)
        self._value = value

    @property
    def value(self) -> float:
        """Provides the value.

        Returns:
            The float value
        """
        return self._value

    def accept(self, visitor: AssertionVisitor) -> None:
        visitor.visit_float_assertion(self)

    def clone(
        self, memo: dict[vr.VariableReference, vr.VariableReference]
    ) -> FloatAssertion:
        return FloatAssertion(self.source.clone(memo), self._value)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, FloatAssertion)
            and self._source == other._source
            and self._value == other._value
        )

    def __hash__(self) -> int:
        return hash((self._source, self._value))


class ObjectAssertion(ReferenceAssertion):
    """An assertion on the object behind a reference.

    This can be anything where the object can be reliably copied, besides float, which
    is handled by FloatAssertion.

    For example:
        assert var_0 == [1,2,3]
        assert var_1 == {Foo.BAR}
        assert var_2 == "Foobar"
    """

    def __init__(self, source: vr.Reference, value: Any):
        super().__init__(source)
        self._object = value

    @property
    def object(self) -> Any:
        """Provides the object used for comparison.

        Returns:
            The object used for comparison.
        """
        return self._object

    def accept(self, visitor: AssertionVisitor) -> None:
        visitor.visit_object_assertion(self)

    def clone(
        self, memo: dict[vr.VariableReference, vr.VariableReference]
    ) -> ObjectAssertion:
        return ObjectAssertion(self.source.clone(memo), self._object)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ObjectAssertion)
            and self._source == other._source
            and self._object == other._object
        )

    def __hash__(self) -> int:
        # We cannot include the value in the hash, as it might be unhashable.
        # For example, dicts, lists and sets are not hashable.
        return 17 * hash(self._source) + 31


class CollectionLengthAssertion(ReferenceAssertion):
    """An assertion on the length of a reference.

    This can be necessary for a collection that contains elements that we can not
    directly assert on. Though this kind of assertion is less preferable because it is
    less precise.

    For example:
        assert len(var_0) = 42
    """

    def __init__(self, source: vr.Reference, length: int):
        super().__init__(source)
        self._length = length

    @property
    def length(self) -> Any:
        """The expected length.

        Returns:
            The expected length.
        """
        return self._length

    def accept(self, visitor: AssertionVisitor) -> None:
        visitor.visit_collection_length_assertion(self)

    def clone(
        self, memo: dict[vr.VariableReference, vr.VariableReference]
    ) -> CollectionLengthAssertion:
        return CollectionLengthAssertion(self.source.clone(memo), self._length)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, CollectionLengthAssertion)
            and self._source == other._source
            and self._length == other._length
        )

    def __hash__(self) -> int:
        return hash((self._source, self._length))


class AssertionVisitor:
    """Abstract visitor for assertions."""

    @abstractmethod
    def visit_not_none_assertion(self, assertion: NotNoneAssertion) -> None:
        """Visit a none assertion.

        Args:
            assertion: the visited assertion

        """

    @abstractmethod
    def visit_float_assertion(self, assertion: FloatAssertion) -> None:
        """Visit a float assertion.

        Args:
            assertion: the visited assertion

        """

    @abstractmethod
    def visit_object_assertion(self, assertion: ObjectAssertion) -> None:
        """Visit an object assertion.

        Args:
            assertion: the visited assertion

        """

    @abstractmethod
    def visit_collection_length_assertion(
        self, assertion: CollectionLengthAssertion
    ) -> None:
        """Visit a collection length assertion.

        Args:
            assertion: the visited assertion

        """



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides analyses implemented on the abstract syntax tree.

The implementation of this module contains some code adopted from the ``darglint``
library (https://github.com/terrencepreilly/darglint), which was released by Terrence
Reilly under MIT license.
"""
from __future__ import annotations

import ast
import dataclasses
import enum
import logging
from collections import deque
from typing import Any, Iterable, Iterator

_LOGGER = logging.getLogger(__name__)


def has_decorator(
    func: ast.FunctionDef | ast.AsyncFunctionDef,
    decorators: str | Iterable[str],
) -> bool:
    """Checks whether a function has one or more decorators.

    Args:
        func: The function node from the AST
        decorators: The name or a list of names of decorators to check

    Returns:
        Whether or not the function has the decorators
    """
    if isinstance(decorators, str):
        decorators = (decorators,)

    for decorator in func.decorator_list:
        if isinstance(decorator, ast.Name) and decorator.id in decorators:
            return True
    return False


def get_docstring(node: ast.AST) -> str | None:
    """Retrieves the docstring for an AST node if any.

    If the node does not provide a docstring, it raises a ``TypeError``.

    Args:
        node: The AST node

    Returns:
        The docstring for that node, if any
    """
    return ast.get_docstring(node)


def get_all_functions(
    tree: ast.AST,
) -> Iterator[ast.FunctionDef | ast.AsyncFunctionDef]:
    """Yields all functions from an AST.

    Args:
        tree: The AST

    Yields:
        All functions from the AST
    """
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            yield node


def get_all_classes(tree: ast.AST) -> Iterator[ast.ClassDef]:
    """Yields all classes from an AST.

    Args:
        tree: The AST

    Yields:
        All classes from the AST
    """
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            yield node


def get_all_methods(
    tree: ast.AST,
) -> Iterator[ast.FunctionDef | ast.AsyncFunctionDef]:
    """Yields all methods from an AST.

    Args:
        tree: The AST

    Yields:
        All methods from the AST
    """
    for class_ in get_all_classes(tree):
        yield from get_all_functions(class_)


def get_return_type(func: ast.FunctionDef | ast.AsyncFunctionDef) -> str | None:
    """Retrieves the return type of a function from the AST.

    Args:
        func: The function-definition node from the AST

    Returns:
        The function's return type, might be ``None``
    """
    if func.returns is not None and hasattr(func.returns, "id"):
        return getattr(func.returns, "id")
    return None


def get_line_number_for_function(func: ast.FunctionDef | ast.AsyncFunctionDef) -> int:
    """Retrieves the line number for a function from the AST.

    Args:
        func: The function-definition node from the AST

    Returns:
        The function's line number
    """
    line_number = func.lineno
    if hasattr(func, "args") and func.args.args:
        last_arg = func.args.args[-1]
        line_number = last_arg.lineno
    return line_number


class FunctionAndMethodVisitor(ast.NodeVisitor):
    """Extracts all functions, methods, and properties from an AST."""

    def __init__(self) -> None:
        self.__callables: set[ast.FunctionDef | ast.AsyncFunctionDef] = set()
        self.__methods: set[ast.FunctionDef | ast.AsyncFunctionDef] = set()
        self.__properties: set[ast.FunctionDef | ast.AsyncFunctionDef] = set()

    @property
    def functions(self) -> list[ast.FunctionDef | ast.AsyncFunctionDef]:
        """Provides all traced functions."""
        return list(self.__callables - self.__methods - self.__properties)

    @property
    def methods(self) -> list[ast.FunctionDef | ast.AsyncFunctionDef]:
        """Provides all traced methods."""
        return list(self.__methods)

    @property
    def properties(self) -> list[ast.FunctionDef | ast.AsyncFunctionDef]:
        """Provides all traced properties."""
        return list(self.__properties)

    # pylint: disable=invalid-name, missing-docstring
    def visit_ClassDef(self, node: ast.ClassDef) -> ast.AST:
        for item in node.body:
            if isinstance(item, (ast.AsyncFunctionDef, ast.FunctionDef)):
                if has_decorator(item, "property"):
                    self.__properties.add(item)
                else:
                    self.__methods.add(item)
        return self.generic_visit(node)

    # pylint: disable=invalid-name, missing-docstring
    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        self.__callables.add(node)
        return self.generic_visit(node)

    # pylint: disable=invalid-name, missing-docstring
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> ast.AST:
        self.__callables.add(node)
        return self.generic_visit(node)


class _FunctionScopedVisitorMixin(ast.NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.in_function: bool = False

    # pylint: disable=invalid-name, missing-docstring
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> ast.AST:
        if not self.in_function:
            self.in_function = True
            return getattr(super(), "visit_AsyncFunctionDef", super().generic_visit)(
                node
            )
        return ast.Pass()

    # pylint: disable=invalid-name, missing-docstring
    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        if not self.in_function:
            self.in_function = True
            return getattr(super(), "visit_FunctionDef", super().generic_visit)(node)
        return ast.Pass()

    # pylint: disable=invalid-name, missing-docstring
    def visit_Lambda(self, node: ast.Lambda) -> ast.AST:
        if not self.in_function:
            self.in_function = True
            return getattr(super(), "visit_Lambda", super().generic_visit)(node)
        return ast.Pass()


class _AbstractStaticCallableVisitor(ast.NodeVisitor):
    """A visitor that checks for abstract or static methods."""

    def __init__(self) -> None:
        super().__init__()
        self.is_abstract: bool | None = None
        self.is_static: bool = False

    @staticmethod
    def __is_docstring(node: ast.AST) -> bool:
        return (
            isinstance(node, ast.Expr)
            and isinstance(node.value, ast.Constant)
            and isinstance(node.value.value, str)
        )

    @staticmethod
    def __is_ellipsis(node: ast.AST) -> bool:
        return (
            isinstance(node, ast.Expr)
            and isinstance(node.value, ast.Constant)
            and node.value.value is Ellipsis
        )

    # pylint: disable=invalid-name
    @staticmethod
    def __is_raise_NotImplementedException(node: ast.AST) -> bool:
        return isinstance(node, ast.Raise) and (
            (isinstance(node.exc, ast.Name) and node.exc.id == "NotImplementedError")
            or (
                isinstance(node.exc, ast.Call)
                and isinstance(node.exc.func, ast.Name)
                and node.exc.func.id == "NotImplementedError"
            )
        )

    # pylint: disable=invalid-name
    @staticmethod
    def __is_raise_NotImplemented(node: ast.AST) -> bool:
        return (
            isinstance(node, ast.Return)
            and isinstance(node.value, ast.Name)
            and node.value.id == "NotImplemented"
        )

    def __analyse_pure_abstract(
        self, node: ast.AsyncFunctionDef | ast.FunctionDef
    ) -> bool:
        if not has_decorator(node, "abstractmethod"):
            return False

        children = len(node.body)
        if children > 2:
            return False
        if children == 2:
            if not self.__is_docstring(node.body[0]):
                return False

            statement = node.body[1]
        else:
            statement = node.body[0]

        return (
            isinstance(statement, ast.Pass)
            or self.__is_ellipsis(statement)
            or self.__is_raise_NotImplementedException(statement)
            or self.__is_raise_NotImplemented(statement)
            or (children == 1 and self.__is_docstring(statement))
        )

    @staticmethod
    def __analyse_static(node: ast.AsyncFunctionDef | ast.FunctionDef) -> bool:
        return has_decorator(node, "staticmethod")

    # pylint: disable=invalid-name, missing-docstring
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> ast.AST:
        self.is_abstract = self.__analyse_pure_abstract(node)
        self.is_static = self.__analyse_static(node)
        return self.generic_visit(node)

    # pylint: disable=invalid-name, missing-docstring
    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        self.is_abstract = self.__analyse_pure_abstract(node)
        self.is_static = self.__analyse_static(node)
        return self.generic_visit(node)


class _Context:
    """A context tracking exceptions and symbols."""

    def __init__(self) -> None:
        self.exceptions: set[str] = set()

        # If we are in a bare handler, we have to capture new exceptions raised
        # separately from the existing ones.  So, we copy the existing exceptions
        # over here.  This complicates the logic, for the calling class (as
        # contextual operations have to account for two cases), but it does not seem
        # avoidable.
        self.bare_handler_exceptions: set[str] | None = None

        # A lookup from variable names to AST nodes.  If the variable name occurs in
        # a raise exceptions, then the exception will be added using this lookup.
        self.variables: dict[str, str | list[str]] = {}

        # The error(s) which the current exception block is handling.
        self.handling: list[str] | None = None

    def set_in_bare_handler(self) -> None:
        """Move exceptions to bare handler."""
        self.bare_handler_exceptions = set(self.exceptions)
        self.remove_all_exceptions()

    def __get_attr_name(self, attr: ast.Attribute | ast.Name | ast.Tuple) -> list[str]:
        curr: Any = attr
        parts: list[str] = []

        # Assume finite depth of AST here
        while curr:
            if isinstance(curr, ast.Attribute):
                parts.append(curr.attr)
                curr = curr.value
            elif isinstance(curr, ast.Name):
                parts.append(curr.id)
                curr = None
            elif isinstance(curr, ast.Tuple):
                names: list[str] = []
                for node in curr.elts:
                    if isinstance(node, (ast.Attribute, ast.Name)):
                        names.extend(self.__get_attr_name(node))
                    else:
                        _LOGGER.error(
                            "While getting the names from a caught tuple of "
                            "exceptions, encountered something other than an "
                            "ast.Name: %s",
                            node.__class__.__name__,
                        )
                return names
            else:
                _LOGGER.error(
                    "While getting ast.Attribute representation a node had an "
                    "unexpected type %s",
                    curr.__class__.__name__,
                )
                curr = None

        parts.reverse()
        return [".".join(parts)]

    @staticmethod
    def __get_name_name(name: ast.Name | ast.Tuple) -> str | list[str]:
        assert isinstance(name, (ast.Name, ast.Tuple))
        if isinstance(name, ast.Name):
            return name.id
        return [node.id for node in name.elts if isinstance(node, ast.Name)]

    # pylint: disable=too-many-branches, too-many-return-statements
    def __get_exception_name(self, raises: ast.Raise) -> str | list[str]:
        if isinstance(raises, str):
            return raises

        if isinstance(raises.exc, ast.Name):
            name = raises.exc.id
            if name in self.variables:
                return self.variables[name]
            return name
        if isinstance(raises.exc, ast.Call):
            if hasattr(raises.exc.func, "id"):
                return getattr(raises.exc.func, "id")
            if hasattr(raises.exc.func, "attr"):
                return getattr(raises.exc.func, "attr")
            _LOGGER.debug(
                "Raises function call has neither id nor attr, has only %s",
                str(dir(raises.exc.func)),
            )
        elif isinstance(raises.exc, ast.Attribute):
            return raises.exc.attr
        elif isinstance(raises.exc, ast.Subscript):
            id_repr = ""
            if hasattr(raises.exc.value, "id"):
                id_repr = getattr(raises.exc.value, "id")
            n_repr = ""
            if hasattr(raises.exc.slice, "value"):
                value = getattr(raises.exc.slice, "value")
                if hasattr(value, "n"):
                    n_repr = value.n
            return f"{id_repr}[{n_repr}]"
        elif raises.exc is None:
            if not self.handling:
                return ""
            if len(self.handling) == 1:
                return self.handling[0]
            return self.handling
        else:
            _LOGGER.debug("Unexpected type in raises expression: %s", raises.exc)
        return ""

    def add_exception(self, node: ast.Raise) -> set[str]:
        """Add an exception to the context.

        If the exception(s) do not have a name and do not have more children,
        then it is a bare raise.  In that case, we return the exception(s) to the
        parent context.

        Args:
            node: A raise AST node

        Returns:
            A list of exceptions to be passed up to the parent context
        """
        name = self.__get_exception_name(node)
        if name == "":
            if self.bare_handler_exceptions is not None:
                return self.bare_handler_exceptions | self.exceptions
            if self.exceptions:
                return self.exceptions
            if self.variables:
                values: set[str] = set()
                for value in self.variables.values():
                    if isinstance(value, list):
                        values |= set(value)
                    else:
                        values.add(value)
                return values
            _LOGGER.warning(
                "Unexpectedly had no exception name raised and no exception in context."
            )

        if isinstance(name, str):
            self.exceptions.add(name)
        elif isinstance(name, list):
            for part in name:
                self.exceptions.add(part)
        else:
            _LOGGER.warning("Node name extraction failed: %s", node)
        return set()

    def remove_exception(self, node: ast.Raise) -> None:
        """Remove an exception from the context.

        Args:
            node: The raise node
        """
        name = self.__get_exception_name(node)
        if isinstance(name, str) and name in self.exceptions:
            self.exceptions.remove(name)
            self.handling = [name]
        elif isinstance(name, list):
            self.handling = []
            for part in name:
                self.exceptions.remove(part)
                self.handling.append(part)

    def remove_all_exceptions(self) -> None:
        """Removes all exceptions."""
        self.exceptions.clear()

    def add_variable(self, variable: str, exception: ast.Name | ast.Tuple) -> None:
        """Add a variable to the context

        Args:
            variable: The variable
            exception: The bound exception
        """
        self.variables[variable] = self.__get_name_name(exception)

    def set_handling(self, attr: ast.Attribute | ast.Name | ast.Tuple) -> None:
        """Set the handling

        Args:
            attr: An attribute
        """
        self.handling = self.__get_attr_name(attr)

    def remove_variable(self, variable: str) -> None:
        """Remove a variable from this context

        Args:
            variable: The variable
        """
        del self.variables[variable]

    def extend(self, other: _Context) -> None:
        """Merge this context with another context.

        Args:
            other: The other context
        """
        self.exceptions |= other.exceptions

    def finish_handling(self) -> None:
        """Finish the handling of exceptions in this context."""
        self.handling = None


class _RaiseVisitor(ast.NodeVisitor):
    """Inspects raised exceptions to figure out which get exposed outside the func."""

    def __init__(self) -> None:
        super().__init__()
        self.contexts = deque([_Context()])

    @property
    def exceptions(self) -> set[str]:
        """Provides the set of exceptions that are not handled."""
        return self.contexts[0].exceptions

    @property
    def context(self) -> _Context:
        """Provides the outermost context."""
        return self.contexts[-1]

    # pylint: disable=invalid-name, missing-docstring
    def visit_Raise(self, node: ast.Raise) -> ast.AST:
        bubbles = self.context.add_exception(node)
        if bubbles:
            assert len(self.contexts) > 1
            if len(self.contexts) < 2:
                return self.generic_visit(node)
            parent_context = self.contexts[-2]
            parent_context.exceptions |= bubbles

        return self.generic_visit(node)

    # pylint: disable=invalid-name, missing-docstring
    def visit_Try(self, node: ast.Try) -> None:
        self.contexts.append(_Context())
        for child in node.body:
            self.visit(child)
        for handler in node.handlers:
            if handler.type:
                if handler.name and (isinstance(handler.type, (ast.Name, ast.Tuple))):
                    self.context.add_variable(handler.name, handler.type)
                elif isinstance(handler.type, (ast.Attribute, ast.Name, ast.Tuple)):
                    self.context.set_handling(handler.type)
                else:
                    _LOGGER.error(
                        "While getting the types of exceptions in the handler, "
                        "expected to find an ast.Name, ast.Tuple, or ast.Attribute,"
                        "but got %s",
                        handler.type,
                    )
                identifier = getattr(handler.type, "id", None)
                if identifier:
                    self.context.remove_exception(identifier)

                self.generic_visit(handler)
                self.context.finish_handling()
            else:
                # Handle a bare exception.
                #
                # Since the bare exception handles all exceptions, we have to clear
                # all exceptions from the context.  However, exceptions could also be
                # raised from this handler.  So we cannot clear the exceptions first.
                # But if we clear the exceptions second, then remove any new
                # exceptions raised in the handler.  What we need, then, is to know
                # which new exceptions are raised, and clear all but them.  For that,
                # we use a temporary context.
                self.context.set_in_bare_handler()
                self.generic_visit(handler)

        for child in node.finalbody:
            self.visit(child)
        for child in node.orelse:
            self.visit(child)

        context = self.contexts.pop()
        self.context.extend(context)


class _YieldVisitor(ast.NodeVisitor):
    """A visitor checking for ``yield`` nodes."""

    def __init__(self) -> None:
        super().__init__()
        self.yields: list[ast.Yield | ast.YieldFrom] = []

    # pylint: disable=invalid-name, missing-docstring
    def visit_Yield(self, node: ast.Yield) -> ast.AST:
        self.yields.append(node)
        return self.generic_visit(node)

    # pylint: disable=invalid-name, missing-docstring
    def visit_YieldFrom(self, node: ast.YieldFrom) -> ast.AST:
        self.yields.append(node)
        return self.generic_visit(node)


class _ArgumentVisitor(ast.NodeVisitor):
    """Reports the arguments a function contains."""

    def __init__(self) -> None:
        super().__init__()
        self.arguments: list[str] = []
        self.types: list[tuple[str, str | None]] = []

    def __add_arg_by_name(self, name: str, arg: ast.arg) -> None:
        self.arguments.append(name)
        if arg.annotation is not None and hasattr(arg.annotation, "id"):
            # TODO does this cover unions?
            self.types.append((name, getattr(arg.annotation, "id")))
        else:
            self.types.append((name, None))

    # pylint: disable=missing-docstring
    def visit_arguments(self, node: ast.arguments) -> ast.AST:
        for arg in node.posonlyargs:
            self.__add_arg_by_name(arg.arg, arg)

        for arg in node.args:
            self.__add_arg_by_name(arg.arg, arg)

        for arg in node.kwonlyargs:
            self.__add_arg_by_name(arg.arg, arg)

        if node.vararg is not None:
            name = f"*{node.vararg.arg}"
            self.__add_arg_by_name(name, node.vararg)

        if node.kwarg is not None:
            name = f"**{node.kwarg.arg}"
            self.__add_arg_by_name(name, node.kwarg)

        return self.generic_visit(node)


class _VariableVisitor(ast.NodeVisitor):
    """Collect variables."""

    def __init__(self) -> None:
        super().__init__()
        self.variables: list[ast.Name] = []

    # pylint: disable=invalid-name, missing-docstring
    def visit_Name(self, node: ast.Name) -> ast.AST:
        if hasattr(node, "ctx") and isinstance(node.ctx, ast.Store):
            self.variables.append(node)
        return self.generic_visit(node)


class _ReturnVisitor(ast.NodeVisitor):
    """A visitor checking for ``return`` nodes"""

    def __init__(self) -> None:
        super().__init__()
        self.returns: list[ast.Return | None] = []
        self.return_types: list[ast.AST | None] = []

    # pylint: disable=invalid-name, missing-docstring
    def visit_Return(self, node: ast.Return) -> ast.AST:
        self.returns.append(node)
        return self.generic_visit(node)


class _AssertVisitor(ast.NodeVisitor):
    """A visitor checking for ``assert`` statements."""

    def __init__(self) -> None:
        super().__init__()
        self.asserts: list[ast.Assert] = []

    # pylint: disable=invalid-name, missing-docstring
    def visit_Assert(self, node: ast.Assert) -> ast.AST:
        self.asserts.append(node)
        return self.generic_visit(node)


# pylint: disable=too-many-ancestors
class FunctionAnalysisVisitor(
    _FunctionScopedVisitorMixin,  # needs to be first in order!
    _AbstractStaticCallableVisitor,
    _RaiseVisitor,
    _YieldVisitor,
    _ArgumentVisitor,
    _VariableVisitor,
    _ReturnVisitor,
    _AssertVisitor,
):
    """A visitor that analyses functions.

    It assumes that it will be only called on ``ast.FunctionDef`` or
    ``ast.AsyncFunctionDef`` nodes.
    """


class FunctionType(enum.Enum):
    """Specifies the type of a function."""

    FUNCTION = enum.auto()
    """Denotes a function, i.e., a callable on the top level of the module."""

    METHOD = enum.auto()
    """Denotes a method, i.e., a callable that is part of a class."""

    PROPERTY = enum.auto()
    """Denotes a property, i.e., a callable that has the ``@property`` decorator."""


@dataclasses.dataclass
class FunctionDescription:  # pylint: disable=too-many-instance-attributes
    """Describes a function or method in the subject under test.

    Attributes:
        argument_names: The (potentially empty) list of arguments of the function
        argument_types: The list of arguments and their annotated type (or ``None``)
        docstring: The optional docstring of the function
        func: The AST node of the function
        has_empty_return: Whether the function has an empty ``return`` statement
        has_return: Whether the function has a ``return`` statement
        has_yield: Whether there is a ``yield`` statement in the function's body
        is_abstract: Whether the function is an abstract method of a class
        is_method: Whether the function is a method of a class
        is_property: Whether the function is a property of a class
        is_static: Whether the function is a static method of a class
        line_number: The line number the function is defined in
        name: The name of the function
        raises: The (potentially empty) set of exceptions the function raises
        raises_assert: Whether the function raises any exceptions
        return_type: The annotated type the function returns (if any)
        return_value: The return node from the AST (if any)
        variables: A list of variables that get defined inside the function
    """

    argument_names: list[str]
    argument_types: list[tuple[str, str | None]]
    docstring: str | None
    func: ast.AsyncFunctionDef | ast.FunctionDef
    has_empty_return: bool
    has_return: bool
    has_yield: bool
    is_abstract: bool | None
    is_method: bool
    is_property: bool
    is_static: bool
    line_number: int
    name: str
    raises: set[str]
    raises_assert: bool
    return_type: str | None
    return_value: ast.Return | None
    variables: list[ast.Name]


def get_function_descriptions(program: ast.AST) -> list[FunctionDescription]:
    """Extracts the function descriptions from the AST.

    Args:
        program: The program's AST

    Returns:
        A list of function descriptions extracted from the AST
    """
    result: list[FunctionDescription] = []
    functions_methods = FunctionAndMethodVisitor()
    functions_methods.visit(program)
    for prop in functions_methods.properties:
        result.append(
            __build_function_description(function_type=FunctionType.PROPERTY, func=prop)
        )
    for method in functions_methods.methods:
        result.append(
            __build_function_description(function_type=FunctionType.METHOD, func=method)
        )
    for function in functions_methods.functions:
        result.append(
            __build_function_description(
                function_type=FunctionType.FUNCTION, func=function
            )
        )
    return result


def __build_function_description(
    function_type: FunctionType, func: ast.AsyncFunctionDef | ast.FunctionDef
) -> FunctionDescription:
    function_analysis = FunctionAnalysisVisitor()
    function_analysis.visit(func)

    arguments = function_analysis.arguments
    argument_types = function_analysis.types
    if function_type != FunctionType.FUNCTION and len(arguments) > 0:
        if not has_decorator(func, "staticmethod"):
            arguments.pop(0)
            argument_types.pop(0)

    has_return = bool(function_analysis.returns)
    has_empty_return = False
    return_value = None
    if has_return:
        return_value = function_analysis.returns[0]
        has_empty_return = return_value is not None and return_value.value is None

    return FunctionDescription(
        argument_names=arguments,
        argument_types=argument_types,
        docstring=get_docstring(func),
        func=func,
        has_empty_return=has_empty_return,
        has_return=has_return,
        has_yield=bool(function_analysis.yields),
        is_abstract=function_analysis.is_abstract,
        is_method=(function_type == FunctionType.METHOD),
        is_property=(function_type == FunctionType.PROPERTY),
        is_static=function_analysis.is_static,
        line_number=get_line_number_for_function(func),
        name=func.name,
        raises=function_analysis.exceptions,
        raises_assert=bool(function_analysis.asserts),
        return_type=get_return_type(func),
        return_value=return_value,
        variables=function_analysis.variables,
    )



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides various crowding-distance assignment implementations."""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING, TypeVar

from ordered_set import OrderedSet

import pynguin.ga.chromosome as chrom

if TYPE_CHECKING:
    import pynguin.ga.computations as ff

C = TypeVar("C", bound=chrom.Chromosome)  # pylint: disable=invalid-name


def fast_epsilon_dominance_assignment(
    front: list[C], goals: OrderedSet[ff.FitnessFunction]
) -> None:
    """Implements a "fast" version of the variant of the crowding distance.

    It is named "epsilon-dominance-assignment" and was proposed by Köppen and Yoshida in
    Mario Köppen and Kaori Yoshida, "Substitute Distance Assignments in NSGA-II for
    handling Many-objective Optimization Problems", Evolutionary Multi-Croterion
    Optimization, LNCS vol. 4403, 2007, pp. 727–741.

    Args:
        front: Front of non-dominated solutions/tests
        goals: Set of goals/targets (e.g., branches) to consider
    """
    for test in front:
        test.distance = 0

    for goal in goals:
        minimum = sys.float_info.max
        min_set: list[C] = []
        maximum = 0.0
        for test in front:
            value = test.get_fitness_for(goal)
            if value < minimum:
                minimum = value
                min_set.clear()
                min_set.append(test)
            elif value == minimum:
                min_set.append(test)

            if value > maximum:
                maximum = value

        if maximum == minimum:
            continue

        for test in min_set:
            numerator = len(front) - len(min_set)
            denominator = len(front)
            test.distance = max(test.distance, numerator / denominator)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an implementation of a dominator tree."""
from __future__ import annotations

import queue
from typing import TYPE_CHECKING

import pynguin.analyses.controlflow.programgraph as pg

if TYPE_CHECKING:
    from pynguin.analyses.controlflow import cfg


class DominatorTree(pg.ProgramGraph[pg.ProgramGraphNode]):
    """Implements a dominator tree."""

    @staticmethod
    def compute(graph: cfg.CFG) -> DominatorTree:
        """Computes the dominator tree for a control-flow graph.

        Args:
            graph: The control-flow graph

        Returns:
            The dominator tree for the control-flow graph
        """
        return DominatorTree.compute_dominance_tree(graph)

    @staticmethod
    def compute_post_dominator_tree(graph: cfg.CFG) -> DominatorTree:
        """Computes the post-dominator tree for a control-flow graph.

        Args:
            graph: The control-flow graph

        Returns:
            The post-dominator tree for the control-flow graph
        """
        reversed_cfg = graph.reversed()
        return DominatorTree.compute(reversed_cfg)

    @staticmethod
    def compute_dominance_tree(graph: cfg.CFG) -> DominatorTree:
        """Computes the dominance tree for a control-flow graph.

        Args:
            graph: The control-flow graph

        Returns:
            The dominance tree for the control-flow graph
        """
        dominance: dict[
            pg.ProgramGraphNode, set[pg.ProgramGraphNode]
        ] = DominatorTree._calculate_dominance(graph)
        for dominance_node, nodes in dominance.items():
            nodes.discard(dominance_node)
        dominance_tree = DominatorTree()
        entry_node = graph.entry_node
        assert entry_node is not None
        dominance_tree.add_node(entry_node)

        node_queue: queue.SimpleQueue = queue.SimpleQueue()
        node_queue.put(entry_node)
        while not node_queue.empty():
            node: pg.ProgramGraphNode = node_queue.get()
            for current, dominators in dominance.items():
                if node in dominators:
                    dominators.remove(node)
                    if len(dominators) == 0:
                        dominance_tree.add_node(current)
                        dominance_tree.add_edge(node, current)
                        node_queue.put(current)
        return dominance_tree

    @staticmethod
    def _calculate_dominance(
        graph: cfg.CFG,
    ) -> dict[pg.ProgramGraphNode, set[pg.ProgramGraphNode]]:
        dominance_map: dict[pg.ProgramGraphNode, set[pg.ProgramGraphNode]] = {}
        entry = graph.entry_node
        assert entry, "Cannot work with a graph without entry nodes"
        entry_dominators: set[pg.ProgramGraphNode] = {entry}
        dominance_map[entry] = entry_dominators

        for node in graph.nodes:
            if node == entry:
                continue
            all_nodes: set[pg.ProgramGraphNode] = set(graph.nodes)
            dominance_map[node] = all_nodes

        changed: bool = True
        while changed:
            changed = False
            for node in graph.nodes:
                if node == entry:
                    continue
                current_dominators = dominance_map.get(node)
                new_dominators = DominatorTree._calculate_dominators(
                    graph, dominance_map, node
                )

                if current_dominators != new_dominators:
                    changed = True
                    dominance_map[node] = new_dominators
                    break

        return dominance_map

    @staticmethod
    def _calculate_dominators(
        graph: cfg.CFG,
        dominance_map: dict[pg.ProgramGraphNode, set[pg.ProgramGraphNode]],
        node: pg.ProgramGraphNode,
    ) -> set[pg.ProgramGraphNode]:
        dominators: set[pg.ProgramGraphNode] = {node}
        intersection: set[pg.ProgramGraphNode] = set()
        predecessors = graph.get_predecessors(node)
        if not predecessors:
            return set()

        first_time: bool = True
        for predecessor in predecessors:
            predecessor_dominators = dominance_map.get(predecessor)
            assert predecessor_dominators is not None, "Cannot be None"
            if first_time:
                intersection = intersection.union(predecessor_dominators)
                first_time = False
            else:
                intersection.intersection_update(predecessor_dominators)
        intersection = intersection.union(dominators)
        return intersection



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""
Provide wrappers around constructors, methods, function, fields and enums.
Think of these like the reflection classes in Java.
"""
from __future__ import annotations

import abc
import enum
import typing
from typing import Callable, List, cast

if typing.TYPE_CHECKING:
    from pynguin.typeinference.strategy import InferredSignature


class GenericAccessibleObject(metaclass=abc.ABCMeta):
    """Abstract base class for something that can be accessed."""

    def __init__(self, owner: type | None):
        self._owner = owner

    @abc.abstractmethod
    def generated_type(self) -> type | None:
        """Provides the type that is generated by this accessible object.

        Returns:
            The generated type  # noqa: DAR202
        """

    @property
    def owner(self) -> type | None:
        """The type which owns this accessible object.

        Returns:
            The owner of this accessible object
        """
        return self._owner

    # pylint: disable=no-self-use
    def is_enum(self) -> bool:
        """Is this a enum?

        Returns:
            Whether or not this is an enum
        """
        return False

    # pylint: disable=no-self-use
    def is_method(self) -> bool:
        """Is this a method?

        Returns:
            Whether or not this is a method
        """
        return False

    # pylint: disable=no-self-use
    def is_constructor(self) -> bool:
        """Is this a constructor?

        Returns:
            Whether or not this is a constructor
        """
        return False

    # pylint: disable=no-self-use
    def is_function(self) -> bool:
        """Is this a function?

        Returns:
            Whether or not this is a function
        """
        return False

    # pylint: disable=no-self-use
    def is_field(self) -> bool:
        """Is this a field?

        Returns:
            Whether or not this is a field
        """
        return False

    # pylint: disable=no-self-use
    def is_static(self) -> bool:
        """Is this static?

        Returns:
            Whether or not this is static
        """
        return False

    # pylint: disable=no-self-use
    def get_num_parameters(self) -> int:
        """Number of parameters.

        Returns:
            The number of parameters
        """
        return 0

    @abc.abstractmethod
    def get_dependencies(self) -> set[type]:
        """A set of types that are required to use this accessible.

        Returns:
            A set of types  # noqa: DAR202
        """


class GenericEnum(GenericAccessibleObject):
    """Models an enum."""

    def __init__(self, owner: type[enum.Enum]):
        super().__init__(owner)
        self._names = list(map(lambda e: e.name, cast(List[enum.Enum], list(owner))))

    def generated_type(self) -> type | None:
        return self._owner

    @property
    def names(self) -> list[str]:
        """All names that this enum has.

        Returns:
            All possible values of this enum."""
        return self._names

    def is_enum(self) -> bool:
        return True

    def get_dependencies(self) -> set[type]:
        return set()  # pylint:disable=no-self-use

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, GenericEnum):
            return False
        return self._owner == other._owner

    def __hash__(self):
        return hash(self._owner)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.owner})"


class GenericCallableAccessibleObject(
    GenericAccessibleObject, metaclass=abc.ABCMeta
):  # pylint: disable=W0223
    """Abstract base class for something that can be called."""

    def __init__(
        self,
        owner: type | None,
        callable_: Callable,
        inferred_signature: InferredSignature,
    ) -> None:
        super().__init__(owner)
        self._callable = callable_
        self._inferred_signature = inferred_signature

    def generated_type(self) -> type | None:
        return self._inferred_signature.return_type

    @property
    def inferred_signature(self) -> InferredSignature:
        """Provides access to the inferred type signature information.

        Returns:
            The inferred type signature
        """
        return self._inferred_signature

    @property
    def callable(self) -> Callable:
        """Provides the callable.

        Returns:
            The callable
        """
        return self._callable

    def get_num_parameters(self) -> int:
        return len(self.inferred_signature.parameters)

    def get_dependencies(self) -> set[type]:
        return {
            value
            for value in self.inferred_signature.parameters.values()
            if value is not None
        }


class GenericConstructor(GenericCallableAccessibleObject):
    """A constructor."""

    def __init__(self, owner: type, inferred_signature: InferredSignature) -> None:
        # super().__init__(owner, owner.__init__, inferred_signature)  # type: ignore
        super().__init__(owner, getattr(owner, "__init__"), inferred_signature)
        assert owner

    def generated_type(self) -> type | None:
        return self.owner

    def is_constructor(self) -> bool:
        return True

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, GenericConstructor):
            return False
        return self._owner == other._owner

    def __hash__(self):
        return hash(self._owner)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.owner}, {self.inferred_signature})"


class GenericMethod(GenericCallableAccessibleObject):
    """A method."""

    def __init__(
        self,
        owner: type,
        method: Callable,
        inferred_signature: InferredSignature,
        method_name: str | None = None,
    ) -> None:
        super().__init__(owner, method, inferred_signature)
        assert owner
        self._method_name = method_name

    @property
    def method_name(self):
        """Returns the name of a generic method.

        Returns:
            The name of a generic method.
        """
        return self._method_name

    def is_method(self) -> bool:
        return True

    def get_dependencies(self) -> set[type]:
        assert self.owner, "Method must have an owner"
        dependencies = super().get_dependencies()
        dependencies.add(self.owner)
        return dependencies

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, GenericMethod):
            return False
        return self._callable == other._callable

    def __hash__(self):
        return hash(self._callable)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.owner},"
            + f" {self._callable.__name__}, {self.inferred_signature})"
        )


class GenericFunction(GenericCallableAccessibleObject):
    """A function, which does not belong to any class."""

    def __init__(
        self,
        function: Callable,
        inferred_signature: InferredSignature,
        function_name: str | None = None,
    ) -> None:
        self._function_name = function_name
        super().__init__(None, function, inferred_signature)

    def is_function(self) -> bool:
        return True

    @property
    def function_name(self) -> str | None:
        """Returns the name of a generic function.

        Returns:
            The name of a generic function.
        """
        return self._function_name

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, GenericFunction):
            return False
        return self._callable == other._callable

    def __hash__(self):
        return hash(self._callable)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self._callable.__name__}, "
            + f"{self.inferred_signature})"
        )


class GenericAbstractField(GenericAccessibleObject, metaclass=abc.ABCMeta):
    """Abstract superclass for fields."""

    def __init__(self, owner: type | None, field: str, field_type: type | None) -> None:
        super().__init__(owner)
        self._field = field
        self._field_type = field_type

    def is_field(self) -> bool:
        return True

    def generated_type(self) -> type | None:
        return self._field_type

    @property
    def field(self) -> str:
        """Provides the name of the field.

        Returns:
            The name of the field
        """
        return self._field


class GenericField(GenericAbstractField):
    """A field of an object."""

    def __init__(self, owner: type, field: str, field_type: type | None):
        super().__init__(owner, field, field_type)
        assert owner, "Field must have an owner"

    def get_dependencies(self) -> set[type]:
        assert self._owner, "Field must have an owner"
        return {self._owner}

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, GenericField):
            return False
        return self._owner == other._owner and self._field == other._field

    def __hash__(self):
        return 31 + 17 * hash(self._owner) + 17 * hash(self._field)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.owner}, {self._field},"
            + f" {self._field_type})"
        )


class GenericStaticField(GenericAbstractField):
    """Static field of a class."""

    def __init__(self, owner: type, field: str, field_type: type | None):
        super().__init__(owner, field, field_type)
        assert owner, "Field must have an owner"

    def is_static(self) -> bool:
        return True

    def get_dependencies(self) -> set[type]:
        return set()

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, GenericStaticField):
            return False
        return self._owner == other._owner and self._field == other._field

    def __hash__(self):
        return 31 + 17 * hash(self._owner) + 17 * hash(self._field)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.owner}, {self._field},"
            + f" {self._field_type})"
        )


class GenericStaticModuleField(GenericAbstractField):
    """Static fields defined in a module."""

    # TODO(fk) combine with regular static field?

    def __init__(self, module: str, field: str, field_type: type | None):
        super().__init__(None, field, field_type)
        self._module = module

    def is_static(self) -> bool:
        return True

    def get_dependencies(self) -> set[type]:
        return set()

    @property
    def module(self) -> str:
        """Provides the name of the module where the field is defined.

        Returns:
            The name of the module where the field is defined.
        """
        return self._module

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, GenericStaticModuleField):
            return False
        return self._module == other._module and self._field == other._field

    def __hash__(self):
        return 31 + 17 * hash(self._module) + 17 * hash(self._field)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self._module}, {self._field},"
            + f" {self._field_type})"
        )



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an adapter for the MutPy mutation testing framework."""
from __future__ import annotations

import logging
from types import ModuleType
from typing import Callable

import mutpy.controller as mc
import mutpy.operators as mo
import mutpy.operators.loop as mol
import mutpy.utils as mu
import mutpy.views as mv

import pynguin.configuration as config
from pynguin.utils.exceptions import ConfigurationException

_LOGGER = logging.getLogger(__name__)


class MutationAdapter:  # pylint: disable=too-few-public-methods
    """Adapter class for interactions with the MutPy mutation testing framework."""

    _strategies: dict[config.MutationStrategy, Callable[[int], mc.HOMStrategy]] = {
        config.MutationStrategy.FIRST_TO_LAST: mc.FirstToLastHOMStrategy,
        config.MutationStrategy.BETWEEN_OPERATORS: mc.BetweenOperatorsHOMStrategy,
        config.MutationStrategy.RANDOM: mc.RandomHOMStrategy,
        config.MutationStrategy.EACH_CHOICE: mc.EachChoiceHOMStrategy,
    }

    def __init__(self):
        self.target_loader: mu.ModulesLoader | None = None

    def mutate_module(self) -> list[tuple[ModuleType, list[mo.Mutation]]]:
        """Mutates the modules specified in the configuration by using MutPys'
        mutation procedure.

        Returns:
            A list of tuples where the first entry is the mutated module and the second
            part is a list of all the mutations operators applied.
        """
        controller = self._build_mutation_controller()
        controller.score = mc.MutationScore()

        mutants = []

        if self.target_loader is not None:
            for target_module, to_mutate in self.target_loader.load():
                _LOGGER.info("Build AST for %s", target_module.__name__)
                target_ast = controller.create_target_ast(target_module)
                _LOGGER.info("Mutate module %s", target_module.__name__)
                mutant_modules = controller.mutate_module(
                    target_module=target_module,
                    to_mutate=to_mutate,
                    target_ast=target_ast,
                )
                for mutant_module, mutations in mutant_modules:
                    mutants.append((mutant_module, mutations))
        _LOGGER.info("Generated %d mutants", len(mutants))
        return mutants

    def _build_mutation_controller(self) -> mc.MutationController:
        _LOGGER.info("Setup mutation controller")
        built_views = self._get_views()
        mutant_generator = self._get_mutant_generator()
        self.target_loader = mu.ModulesLoader(
            [config.configuration.module_name], config.configuration.project_path
        )

        return mc.MutationController(
            runner_cls=None,
            target_loader=self.target_loader,
            test_loader=None,
            views=built_views,
            mutant_generator=mutant_generator,
        )

    def _get_mutant_generator(self) -> mc.FirstOrderMutator:
        operators_set = set()
        operators_set |= mo.standard_operators

        # Only use a selected set of the experimental operators.
        operators_set |= {
            mol.OneIterationLoop,
            mol.ReverseIterationLoop,
            mol.ZeroIterationLoop,
        }

        # percentage of the generated mutants (mutation sampling)
        percentage = 100

        mutation_strategy = config.configuration.test_case_output.mutation_strategy

        if mutation_strategy == config.MutationStrategy.FIRST_ORDER_MUTANTS:
            return mc.FirstOrderMutator(operators_set, percentage)

        order = config.configuration.test_case_output.mutation_order
        if order <= 0:
            raise ConfigurationException("Mutation order should be > 0.")

        if mutation_strategy in self._strategies:
            hom_strategy = self._strategies[mutation_strategy](order)
            return mc.HighOrderMutator(operators_set, percentage, hom_strategy)
        raise ConfigurationException("No suitable mutation strategy found.")

    @staticmethod
    def _get_views() -> list[mv.QuietTextView]:
        # We do not want any output from MutPy here
        return [mv.QuietTextView()]



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a converter to unify generation results."""
from __future__ import annotations

from typing import TYPE_CHECKING

import pynguin.ga.chromosomevisitor as cv
import pynguin.ga.testsuitechromosome as tsc

if TYPE_CHECKING:
    import pynguin.ga.testcasechromosome as tcc


class ChromosomeConverter(cv.ChromosomeVisitor):
    """
    A chromosome visitor that collects any chromosomes and partitions them
    into a failing and passing test suite chromosome.

    The passing test suite contains only those test cases that did not raise an
    exception during execution.  The failing test suite the remaining test cases.
    TODO(sl) Take care for test cases that raise an exception on purpose.
    """

    def __init__(self):
        """Create new chromosome converter."""
        # TODO(fk) Need to handle fitness functions.
        self._failing_test_suite = tsc.TestSuiteChromosome()
        self._passing_test_suite = tsc.TestSuiteChromosome()

    @property
    def passing_test_suite(self) -> tsc.TestSuiteChromosome:
        """Provides the test suite chromosome containing the passing test cases

        Returns:
            the test suite with the passing test cases
        """
        return self._passing_test_suite

    @property
    def failing_test_suite(self) -> tsc.TestSuiteChromosome:
        """Provides the test suite chromosome containing the failing test cases

        Returns:
            the test suite with the failing test cases
        """
        return self._failing_test_suite

    def visit_test_suite_chromosome(self, chromosome: tsc.TestSuiteChromosome) -> None:
        for test_case_chromosome in chromosome.test_case_chromosomes:
            test_case_chromosome.accept(self)

    def visit_test_case_chromosome(self, chromosome: tcc.TestCaseChromosome) -> None:
        if chromosome.is_failing():
            self._failing_test_suite.add_test_case_chromosome(chromosome.clone())
        else:
            self._passing_test_suite.add_test_case_chromosome(chromosome.clone())



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides implementations of a ranking function."""
from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, TypeVar

from ordered_set import OrderedSet

import pynguin.configuration as config
import pynguin.ga.chromosome as chrom
from pynguin.ga.comparators.dominancecomparator import DominanceComparator
from pynguin.ga.comparators.preferencesortingcomparator import (
    PreferenceSortingComparator,
)
from pynguin.utils import randomness

if TYPE_CHECKING:
    import pynguin.ga.computations as ff

C = TypeVar("C", bound=chrom.Chromosome)  # pylint: disable=invalid-name


@dataclass
class RankedFronts(Generic[C]):
    """Contains the ranked fronts."""

    fronts: list[list[C]] | None = None

    def get_sub_front(self, rank: int) -> list[C]:
        """Returns the sub-front of chromosome objects of the given rank.

        Sub-fronts are ordered starting from 0 in ascending order, i.e., the first
        non-dominated front has rank 0, the next sub-front rank 1 etc.

        Args:
            rank: The sub-front to retrieve

        Returns:
            A list of solutions of a given rank
        """
        if self.fronts is None or rank >= len(self.fronts):
            return []
        return self.fronts[rank]

    def get_number_of_sub_fronts(self) -> int:
        """Returns the total number of sub-fronts found.

        Returns:
            The total number of sub-fronts found
        """
        assert self.fronts is not None
        return len(self.fronts)


# pylint: disable=too-few-public-methods
class RankingFunction(Generic[C], metaclass=ABCMeta):
    """Interface for ranking algorithms."""

    @abstractmethod
    def compute_ranking_assignment(
        self, solutions: list[C], uncovered_goals: OrderedSet[ff.FitnessFunction]
    ) -> RankedFronts:
        """Computes the ranking assignment for the given population of solutions.

        The computation is done with respect to the given set of coverage goals.
        More precisely, every individual in the population is assigned to a specific
        dominance front, which can afterwards be retrieved by calling `get_sub_front(
        int)`.  The concrete dominance comparator used for computing the ranking is
        defined by subclasses implementing this interface.

        Args:
            solutions: The population to rank
            uncovered_goals: The set of coverage goals to consider for the ranking
                             assignment

        Returns:
            The ranked fronts
        """


# pylint: disable=too-few-public-methods
class RankBasedPreferenceSorting(RankingFunction, Generic[C]):
    """Ranks the test cases according to the preference criterion defined for MOSA."""

    _logger = logging.getLogger(__name__)

    def compute_ranking_assignment(
        self, solutions: list[C], uncovered_goals: OrderedSet[ff.FitnessFunction]
    ) -> RankedFronts:
        if not solutions:
            self._logger.debug("Solution is empty")
            return RankedFronts()

        fronts = []

        # First apply the "preference sorting" to the first front only then compute
        # the ranks according to the non-dominate sorting algorithm
        zero_front: list[C] = self._get_zero_front(solutions, uncovered_goals)
        fronts.append(zero_front)
        front_index = 1

        if len(zero_front) < config.configuration.search_algorithm.population:
            ranked_solutions = len(zero_front)
            comparator: DominanceComparator[C] = DominanceComparator(
                goals=uncovered_goals
            )

            remaining: list[C] = []
            remaining.extend(solutions)
            for element in zero_front:
                if element in remaining:
                    remaining.remove(element)

            while (
                ranked_solutions < config.configuration.search_algorithm.population
                and len(remaining) > 0
            ):
                new_front: list[C] = self._get_non_dominated_solutions(
                    remaining, comparator, front_index
                )
                fronts.append(new_front)
                for element in new_front:
                    if element in remaining:
                        remaining.remove(element)
                ranked_solutions += len(new_front)
                front_index += 1

        else:
            remaining = []
            remaining.extend(solutions)
            for element in zero_front:
                if element in remaining:
                    remaining.remove(element)
            for element in remaining:
                element.rank = front_index
            fronts.append(remaining)

        return RankedFronts(fronts)

    @staticmethod
    def _get_zero_front(
        solutions: list[C], uncovered_goals: OrderedSet[ff.FitnessFunction]
    ) -> list[C]:
        zero_front: OrderedSet[C] = OrderedSet()
        for goal in uncovered_goals:
            comparator: PreferenceSortingComparator[C] = PreferenceSortingComparator(
                goal
            )
            best: C | None = None
            for solution in solutions:
                flag = comparator.compare(solution, best)
                if flag < 0 or (flag == 0 and randomness.next_bool()):
                    best = solution
            assert best is not None

            best.rank = 0
            zero_front.add(best)
        return list(zero_front)

    @staticmethod
    def _get_non_dominated_solutions(
        solutions: list[C], comparator: DominanceComparator[C], front_index: int
    ) -> list[C]:
        front: list[C] = []
        for solution in solutions:
            is_dominated = False
            dominated_solutions: list[C] = []
            for best in front:
                flag = comparator.compare(solution, best)
                if flag < 0:
                    dominated_solutions.append(best)
                if flag > 0:
                    is_dominated = True
                    break
            if is_dominated:
                continue

            solution.rank = front_index
            front.append(solution)
            for dominated_solution in dominated_solutions:
                if dominated_solution in front:
                    front.remove(dominated_solution)
        return front



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provide rank selection."""
from __future__ import annotations

from math import sqrt
from typing import TypeVar

import pynguin.configuration as config
import pynguin.ga.chromosome as chrom
from pynguin.ga.operators.selection.selection import SelectionFunction
from pynguin.utils import randomness

T = TypeVar("T", bound=chrom.Chromosome)  # pylint:disable=invalid-name


class RankSelection(SelectionFunction[T]):
    """Rank selection."""

    def get_index(self, population: list[T]) -> int:
        """Provides an index in the population that is chosen by rank selection.

        Make sure that the population is sorted. The fittest chromosomes have to
        come first.

        Args:
            population: A list of chromosomes to select from

        Returns:
            The index that should be used for selection
        """
        random_value = randomness.next_float()
        bias = config.configuration.search_algorithm.rank_bias
        return int(
            len(population)
            * (
                (bias - sqrt(bias ** 2 - (4.0 * (bias - 1.0) * random_value)))
                / 2.0
                / (bias - 1.0)
            )
        )
